Type: Package
Package: wbwdi
Title: Seamless Access to World Bank World Development Indicators (WDI)
Version: 1.0.4
Authors@R: c(
    person("Christoph", "Scheuch", , "christoph@tidy-intelligence.com", 
            role = c("aut", "cre", "cph"),
            comment = c(ORCID = "0009-0004-0423-6819"))
    )
Description: Access and analyze the World Bank's World Development Indicators 
  (WDI) using the corresponding API <https://datahelpdesk.worldbank.org/knowledgebase/articles/889392-about-the-indicators-api-documentation>. 
  WDI provides more than 24,000 country or region-level indicators for various
  contexts. 'wbwdi' enables users to download, process and work with WDI
  series across multiple countries, aggregates, and time periods. 
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1)
Imports: cli, dplyr (>= 1.0.0), httr2 (>= 1.0.0), jsonlite (>= 1.0.0),
        purrr (>= 1.0.0), tibble, tidyr (>= 1.0.0), rlang (>= 1.0.0)
Suggests: curl, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/Needs/website: forcats, ggplot2, wbids
URL: https://github.com/tidy-intelligence/r-wbwdi,
        https://tidy-intelligence.github.io/r-wbwdi/
BugReports: https://github.com/tidy-intelligence/r-wbwdi/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-10 13:24:43 UTC; krise
Author: Christoph Scheuch [aut, cre, cph] (ORCID:
    <https://orcid.org/0009-0004-0423-6819>)
Maintainer: Christoph Scheuch <christoph@tidy-intelligence.com>
Repository: CRAN
Date/Publication: 2026-04-10 16:50:02 UTC
Built: R 4.5.2; ; 2026-04-15 05:12:07 UTC; unix

Type: Package
Package: wbwdi
Title: Seamless Access to World Bank World Development Indicators (WDI)
Version: 1.0.4
Authors@R: c(
    person("Christoph", "Scheuch", , "christoph@tidy-intelligence.com", 
            role = c("aut", "cre", "cph"),
            comment = c(ORCID = "0009-0004-0423-6819"))
    )
Description: Access and analyze the World Bank's World Development Indicators 
  (WDI) using the corresponding API <https://datahelpdesk.worldbank.org/knowledgebase/articles/889392-about-the-indicators-api-documentation>. 
  WDI provides more than 24,000 country or region-level indicators for various
  contexts. 'wbwdi' enables users to download, process and work with WDI
  series across multiple countries, aggregates, and time periods. 
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Depends: R (>= 4.1)
Imports: cli, dplyr (>= 1.0.0), httr2 (>= 1.0.0), jsonlite (>= 1.0.0),
        purrr (>= 1.0.0), tibble, tidyr (>= 1.0.0), rlang (>= 1.0.0)
Suggests: curl, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/Needs/website: forcats, ggplot2, wbids
URL: https://github.com/tidy-intelligence/r-wbwdi,
        https://tidy-intelligence.github.io/r-wbwdi/
BugReports: https://github.com/tidy-intelligence/r-wbwdi/issues
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-10 13:24:43 UTC; krise
Author: Christoph Scheuch [aut, cre, cph] (ORCID:
    <https://orcid.org/0009-0004-0423-6819>)
Maintainer: Christoph Scheuch <christoph@tidy-intelligence.com>
Repository: CRAN
Date/Publication: 2026-04-10 16:50:02 UTC
Built: R 4.6.0; ; 2026-04-29 16:07:56 UTC; unix

# Generated by roxygen2: do not edit by hand

export(wdi_get)
export(wdi_get_entities)
export(wdi_get_income_levels)
export(wdi_get_indicators)
export(wdi_get_languages)
export(wdi_get_lending_types)
export(wdi_get_regions)
export(wdi_get_sources)
export(wdi_get_topics)
export(wdi_search)
import(dplyr)
import(httr2)
import(jsonlite)
import(tibble)
importFrom(purrr,map)
importFrom(purrr,map_df)
importFrom(purrr,reduce)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(rlang,eval_tidy)
importFrom(rlang,parse_expr)
importFrom(tidyr,nest)
importFrom(tidyr,pivot_wider)
importFrom(tidyr,unnest)
importFrom(tidyr,unnest_longer)
importFrom(tidyr,unnest_wider)

# py-wbwdi
![PyPI](https://img.shields.io/pypi/v/wbwdi?label=pypi%20package)
![PyPI Downloads](https://img.shields.io/pypi/dm/wbwdi)
[![python-package.yml](https://github.com/tidy-intelligence/py-wbwdi/actions/workflows/python-package.yml/badge.svg)](https://github.com/tidy-intelligence/py-wbwdi/actions/workflows/python-package.yml)
[![codecov.yml](https://codecov.io/gh/tidy-intelligence/py-wbwdi/graph/badge.svg)](https://app.codecov.io/gh/tidy-intelligence/py-wbwdi)
[![License:
MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

`wbwdi` is a Polars-based Python library to access and analyze the World Bank’s World Development Indicators (WDI) using the corresponding [API](https://datahelpdesk.worldbank.org/knowledgebase/articles/889392-about-the-indicators-api-documentation). WDI provides more than 24,000 country or region-level indicators for various contexts. `wbwdi` enables users to download, process and work with WDI series across multiple entities and time periods.

This library is a product of Christoph Scheuch and not sponsored by or affiliated with the World Bank in any way. For an R implementation, please consider the [`r-wbwdi`](https://github.com/tidy-intelligence/r-wbwdi) package. For packages with a shared design philosophy, check out the [econdataverse](https://www.econdataverse.org/).

## Installation

You can install the release version from [PyPI](https://pypi.org/project/wbwdi/): 

```
pip install wbwdi
```

If you want to use the package with `pandas`, then install with the corresponding dependencies:

```
pip install wbwdi[pandas]
```

You also need to set the configuration to `pandas` (see below).

You can install the development version from GitHub:

```
pip install "git+https://github.com/tidy-intelligence/py-wbwdi"
```

## Usage

The main function `wdi_get()` provides an interface to download multiple WDI series for multiple entities and specific date ranges.

```python
import wbwdi as wb

wb.wdi_get(
  entities=["MEX", "CAN", "USA"], 
  indicators=["NY.GDP.PCAP.KD", "SP.POP.TOTL"],
  start_year=2020, 
  end_year=2024
)
```

You can also download these indicators for all entities and available dates:

```python
wb.wdi_get(
  entities="all", 
  indicators=["NY.GDP.PCAP.KD", "SP.POP.TOTL"]
)
```

Some indicators are also available on a monthly basis, e.g.:

```python
wb.wdi_get(
  entities="AUT", 
  indicators="DPANUSSPB",         
  start_year=2012, 
  end_year=2015, 
  frequency="month"
)
```

Similarly, there are also some indicators available on a quarterly frequency, e.g.:

```python
wb.wdi_get(
  entities="NGA", 
  indicators= "DT.DOD.DECT.CD.TL.US",
  start_year=2012, 
  end_year=2015, 
  frequency="quarter"
)
```

You can get a list of all indicators supported by the WDI API via:

```python
wb.wdi_get_indicators()
```

You can get a list of all supported entities via:

```python
wb.wdi_get_entities()
```

You can also get the list of supported indicators and entities in
another language, but note that not everything seems to be translated
into other languages:

```python
wb.wdi_get_indicators(language="es")
wb.wdi_get_entities(language="zh")
```

Check out the following function for a list of supported languages:

```python
wb.wdi_get_languages()
```

In addition, you can list supported regions, sources, topics and lending
types, respectively:

```python
wb.wdi_get_regions()
wb.wdi_get_sources()
wb.wdi_get_topics()
wb.wdi_get_lending_types()
```

If you want to search for specific keywords among indicators or other data sources, you can use the Positron data explorer. Alternatively, this package comes with a helper function:

```python
indicators=wb.wdi_get_indicators()

wb.wdi_search(
  indicators,
  keywords=["inequality", "gender"],
  columns=["indicator_name"]
)
```

If you want to data as a `pandas` or `arrow` data frame instead of `polars`, you can change the corresponding configuration:

```python
wb.wdi_set_format("pandas")  

wb.wdi_get(
  entities=["MEX", "CAN", "USA"], 
  indicators=["NY.GDP.PCAP.KD", "SP.POP.TOTL"],
  start_year=2020, 
  end_year=2024
)
```

## Relation to Existing Python Libraries

There are already great libraries that allow you to interact with the World Bank WDI API. The two main reasons why this library exists are: (i) to have an implementation based on Polars rather than pandas, and (ii) to have an interface consistent with the [econdataverse](https://www.econdataverse.org/).

- [world-bank-data](https://github.com/mwouts/world_bank_data)
- [wbpy](https://github.com/mattduck/wbpy/)
- [wbdata](https://github.com/oliversherouse/wbdata/)
- [pandas_datareader](https://pandas-datareader.readthedocs.io/en/latest/readers/world-bank.html)

wbwdi
================

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/wbwdi)](https://cran.r-project.org/package=wbwdi)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/wbwdi)](https://cran.r-project.org/package=wbwdi)
![R CMD
Check](https://github.com/tidy-intelligence/r-wbwdi/actions/workflows/R-CMD-check.yaml/badge.svg)
![Lint](https://github.com/tidy-intelligence/r-wbwdi/actions/workflows/lint.yaml/badge.svg)
[![Codecov test
coverage](https://codecov.io/gh/tidy-intelligence/r-wbwdi/graph/badge.svg)](https://app.codecov.io/gh/tidy-intelligence/r-wbwdi)
[![License:
MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
<!-- badges: end -->

<!-- README.md is generated from README.Rmd. Please edit that file -->

`wbwdi` is an R package to access and analyze the World Bank’s World
Development Indicators (WDI) using the corresponding
[API](https://datahelpdesk.worldbank.org/knowledgebase/articles/889392-about-the-indicators-api-documentation).
WDI provides more than 24,000 country or region-level indicators for
various contexts. `wbwdi` enables users to download, process and work
with WDI series across multiple entities and time periods.

The package is designed to work seamlessly with International Debt
Statistics (IDS) provided through the
[`wbids`](https://github.com/Teal-Insights/r-wbids) package and shares
its syntax with its sibling Python library
[`wbwdi`](https://github.com/tidy-intelligence/py-wbwdi). It follows the
principles of the [econdataverse](https://www.econdataverse.org/).

This package is a product of Christoph Scheuch and not sponsored by or
affiliated with the World Bank in any way, except for the use of the
World Bank WDI API.

## Installation

You can install `wbwdi` [from
CRAN](https://CRAN.R-project.org/package=wbwdi) via:

``` r
install.packages("wbwdi")
```

You can install the development version of `wbwdi` like this:

``` r
pak::pak("tidy-intelligence/r-wbwdi")
```

## Usage

The main function `wdi_get()` provides an interface to download multiple
WDI series for multiple entities and specific date ranges.

``` r
library(wbwdi)

wdi_get(
  entities = c("MEX", "CAN", "USA"),
  indicators = c("NY.GDP.PCAP.KD", "SP.POP.TOTL"),
  start_year = 2020,
  end_year = 2024
)
#> # A tibble: 30 × 4
#>    entity_id indicator_id    year  value
#>    <chr>     <chr>          <int>  <dbl>
#>  1 CAN       NY.GDP.PCAP.KD  2020 42366.
#>  2 MEX       NY.GDP.PCAP.KD  2020  9235.
#>  3 USA       NY.GDP.PCAP.KD  2020 59195.
#>  4 CAN       NY.GDP.PCAP.KD  2021 44639.
#>  5 MEX       NY.GDP.PCAP.KD  2021  9728.
#>  6 USA       NY.GDP.PCAP.KD  2021 62680.
#>  7 CAN       NY.GDP.PCAP.KD  2022 45678.
#>  8 MEX       NY.GDP.PCAP.KD  2022 10013.
#>  9 USA       NY.GDP.PCAP.KD  2022 63886.
#> 10 CAN       NY.GDP.PCAP.KD  2023 45048.
#> # ℹ 20 more rows
```

You can also download these indicators for all entities and available
dates:

``` r
wdi_get(
  entities = "all",
  indicators = c("NY.GDP.PCAP.KD", "SP.POP.TOTL")
)
#> # A tibble: 35,112 × 4
#>    entity_id indicator_id    year value
#>    <chr>     <chr>          <int> <dbl>
#>  1 AFE       NY.GDP.PCAP.KD  1960 1186.
#>  2 AFW       NY.GDP.PCAP.KD  1960 1130.
#>  3 ARB       NY.GDP.PCAP.KD  1960   NA 
#>  4 CSS       NY.GDP.PCAP.KD  1960 4450.
#>  5 CEB       NY.GDP.PCAP.KD  1960   NA 
#>  6 EAR       NY.GDP.PCAP.KD  1960 1067.
#>  7 EAS       NY.GDP.PCAP.KD  1960 1147.
#>  8 EAP       NY.GDP.PCAP.KD  1960  325.
#>  9 TEA       NY.GDP.PCAP.KD  1960  330.
#> 10 EMU       NY.GDP.PCAP.KD  1960 9952.
#> # ℹ 35,102 more rows
```

Some indicators are also available on a monthly basis, e.g.:

``` r
wdi_get(
  entities = "AUT",
  indicators = "DPANUSSPB",
  start_year = 2012,
  end_year = 2015,
  frequency = "month"
)
#> # A tibble: 48 × 5
#>    entity_id indicator_id  year month value
#>    <chr>     <chr>        <int> <int> <dbl>
#>  1 AUT       DPANUSSPB     2012     1 0.775
#>  2 AUT       DPANUSSPB     2012     2 0.755
#>  3 AUT       DPANUSSPB     2012     3 0.757
#>  4 AUT       DPANUSSPB     2012     4 0.760
#>  5 AUT       DPANUSSPB     2012     5 0.782
#>  6 AUT       DPANUSSPB     2012     6 0.797
#>  7 AUT       DPANUSSPB     2012     7 0.814
#>  8 AUT       DPANUSSPB     2012     8 0.806
#>  9 AUT       DPANUSSPB     2012     9 0.777
#> 10 AUT       DPANUSSPB     2012    10 0.771
#> # ℹ 38 more rows
```

Similarly, there are also some indicators available on a quarterly
frequency, e.g.:

``` r
wdi_get(
  entities = "NGA",
  indicators = "DT.DOD.DECT.CD.TL.US",
  start_year = 2012,
  end_year = 2015,
  frequency = "quarter"
)
#> # A tibble: 16 × 5
#>    entity_id indicator_id          year quarter        value
#>    <chr>     <chr>                <int>   <int>        <dbl>
#>  1 NGA       DT.DOD.DECT.CD.TL.US  2012       1  8235000000 
#>  2 NGA       DT.DOD.DECT.CD.TL.US  2012       2  8374010000 
#>  3 NGA       DT.DOD.DECT.CD.TL.US  2012       3  8587040000 
#>  4 NGA       DT.DOD.DECT.CD.TL.US  2012       4  8883380000 
#>  5 NGA       DT.DOD.DECT.CD.TL.US  2013       1  9195080000 
#>  6 NGA       DT.DOD.DECT.CD.TL.US  2013       2  9477070000 
#>  7 NGA       DT.DOD.DECT.CD.TL.US  2013       3 10817130000 
#>  8 NGA       DT.DOD.DECT.CD.TL.US  2013       4 11402000000 
#>  9 NGA       DT.DOD.DECT.CD.TL.US  2014       1 11746100000 
#> 10 NGA       DT.DOD.DECT.CD.TL.US  2014       2 11957200000 
#> 11 NGA       DT.DOD.DECT.CD.TL.US  2014       3 12002828000 
#> 12 NGA       DT.DOD.DECT.CD.TL.US  2014       4 12138750000 
#> 13 NGA       DT.DOD.DECT.CD.TL.US  2015       1 11781991990.
#> 14 NGA       DT.DOD.DECT.CD.TL.US  2015       2 12667757149.
#> 15 NGA       DT.DOD.DECT.CD.TL.US  2015       3 12966561500 
#> 16 NGA       DT.DOD.DECT.CD.TL.US  2015       4 13040050000
```

You can get a list of all indicators supported by the WDI API via:

``` r
wdi_get_indicators()
#> # A tibble: 29,500 × 7
#>    indicator_id         indicator_name         source_id source_name source_note
#>    <chr>                <chr>                      <int> <chr>       <chr>      
#>  1 1.0.HCount.1.90usd   Poverty Headcount ($1…        37 LAC Equity… "The pover…
#>  2 1.0.HCount.2.5usd    Poverty Headcount ($2…        37 LAC Equity… "The pover…
#>  3 1.0.HCount.Mid10to50 Middle Class ($10-50 …        37 LAC Equity… "The pover…
#>  4 1.0.HCount.Ofcl      Official Moderate Pov…        37 LAC Equity… "The pover…
#>  5 1.0.HCount.Poor4uds  Poverty Headcount ($4…        37 LAC Equity… "The pover…
#>  6 1.0.HCount.Vul4to10  Vulnerable ($4-10 a d…        37 LAC Equity… "The pover…
#>  7 1.0.PGap.1.90usd     Poverty Gap ($1.90 a …        37 LAC Equity… "The pover…
#>  8 1.0.PGap.2.5usd      Poverty Gap ($2.50 a …        37 LAC Equity… "The pover…
#>  9 1.0.PGap.Poor4uds    Poverty Gap ($4 a day)        37 LAC Equity… "The pover…
#> 10 1.0.PSev.1.90usd     Poverty Severity ($1.…        37 LAC Equity… "The pover…
#> # ℹ 29,490 more rows
#> # ℹ 2 more variables: source_organization <chr>, topics <list>
```

You can get a list of all supported entities via:

``` r
wdi_get_entities()
#> # A tibble: 296 × 19
#>    entity_id entity_iso2code entity_type capital_city     entity_name  region_id
#>    <chr>     <chr>           <chr>       <chr>            <chr>        <chr>    
#>  1 ABW       AW              country     Oranjestad       Aruba        LCN      
#>  2 AFE       ZH              aggregates  <NA>             Africa East… NA       
#>  3 AFG       AF              country     Kabul            Afghanistan  MEA      
#>  4 AFR       A9              aggregates  <NA>             Africa       NA       
#>  5 AFW       ZI              aggregates  <NA>             Africa West… NA       
#>  6 AGO       AO              country     Luanda           Angola       SSF      
#>  7 ALB       AL              country     Tirane           Albania      ECS      
#>  8 AND       AD              country     Andorra la Vella Andorra      ECS      
#>  9 ARB       1A              aggregates  <NA>             Arab World   NA       
#> 10 ARE       AE              country     Abu Dhabi        United Arab… MEA      
#> # ℹ 286 more rows
#> # ℹ 13 more variables: region_iso2code <chr>, region_name <chr>,
#> #   admin_region_id <chr>, admin_region_iso2code <chr>,
#> #   admin_region_name <chr>, income_level_id <chr>,
#> #   income_level_iso2code <chr>, income_level_name <chr>,
#> #   lending_type_id <chr>, lending_type_iso2code <chr>,
#> #   lending_type_name <chr>, longitude <dbl>, latitude <dbl>
```

You can also get the list of supported indicators and entities in
another language, but note that not everything seems to be translated
into other languages:

``` r
wdi_get_indicators(language = "es")
wdi_get_entities(language = "zh")
```

Check out the following function for a list of supported languages:

``` r
wdi_get_languages()
#> # A tibble: 23 × 3
#>    language_code language_name native_form     
#>    <chr>         <chr>         <chr>           
#>  1 en            English       English         
#>  2 es            Spanish       Español         
#>  3 fr            French        Français        
#>  4 ar            Arabic        عربي            
#>  5 zh            Chinese       中文            
#>  6 bg            Bulgarian     Български       
#>  7 de            German        Deutsch         
#>  8 hi            Hindi         हिंदी            
#>  9 id            Indonesian    Bahasa indonesia
#> 10 ja            Japanese      日本語          
#> # ℹ 13 more rows
```

In addition, you can list supported regions, sources, topics and lending
types, respectively:

``` r
wdi_get_regions()
wdi_get_sources()
wdi_get_topics()
wdi_get_lending_types()
```

If you want to search for specific keywords among indicators or other
data sources, you can use the RStudio or Positron data explorer.
Alternatively, this package comes with a helper function:

``` r
indicators <- wdi_get_indicators()
wdi_search(
  indicators,
  keywords = c("inequality", "gender"),
  columns = c("indicator_name")
)
#> # A tibble: 476 × 7
#>    indicator_id         indicator_name         source_id source_name source_note
#>    <chr>                <chr>                      <int> <chr>       <chr>      
#>  1 2.3_GIR.GPI          "Gender parity index …        34 Global Par… "Ratio of …
#>  2 2.6_PCR.GPI          "Gender parity index …        34 Global Par… "Ratio of …
#>  3 BI.EMP.PWRK.PB.FE.ZS "Public sector employ…        64 Worldwide …  <NA>      
#>  4 BI.EMP.PWRK.PB.MA.ZS "Public sector employ…        64 Worldwide …  <NA>      
#>  5 BI.EMP.TOTL.PB.FE.ZS "Public sector employ…        64 Worldwide …  <NA>      
#>  6 BI.EMP.TOTL.PB.MA.ZS "Public sector employ…        64 Worldwide …  <NA>      
#>  7 BI.WAG.PREM.PB.FE    "Public sector wage p…        64 Worldwide …  <NA>      
#>  8 BI.WAG.PREM.PB.FM    "P-Value: Public sect…        64 Worldwide …  <NA>      
#>  9 BI.WAG.PREM.PB.FM.CA "P-Value: Gender wage…        64 Worldwide …  <NA>      
#> 10 BI.WAG.PREM.PB.FM.ED "P-Value: Gender wage…        64 Worldwide …  <NA>      
#> # ℹ 466 more rows
#> # ℹ 2 more variables: source_organization <chr>, topics <list>
```

## Relation to Existing R Packages

The most important differences to existing packages are that `wbwdi` is
designed to (i) have a narrow focus on World Bank Development
Indicators, (ii) have a consistent interface with other R packages
(e.g., [`wbids`](https://github.com/Teal-Insights/r-wbids)), (iii) have
an MIT license, and (iv) have a shared interface with Python libraries
(e.g., [`wbwdi`](https://github.com/tidy-intelligence/py-wbwdi)).
`wbwdi` also refrains from using cached data because this approach
frequently leads to problems for users due to outdated caches and it
uses `httr2` to manage API requests and parse responses.

More specifically, the differences of existing CRAN releases (apart from
interface design) are:

- [`WDI`](https://CRAN.R-project.org/package=WDI): uses cached data by
  default, does not allow downloading meta data from the WDI API (e.g.,
  languages, sources, topics), has a GPL-3 license, and does neither use
  `httr` nor `httr2` for requests.
- [`worldbank`](https://CRAN.R-project.org/package=worldbank): does not
  have a narrow focus because it includes the Poverty and Inequality
  Platform and Finances One API.
- [`wbstats`](https://CRAN.R-project.org/package=wbstats): uses cached
  data by default, does not use `httr2`.

## Contributing

Contributions to `econdatasets` are welcome! If you’d like to
contribute, please follow these steps:

1.  **Create an issue**: Before making changes, create an issue
    describing the bug or feature you’re addressing.
2.  **Fork the repository**: After receiving supportive feedback from
    the package authors, fork the repository to your GitHub account.
3.  **Create a branch**: Create a branch for your changes with a
    descriptive name.
4.  **Make your changes**: Implement your bug fix or feature.
5.  **Test your changes**: Run tests to ensure your changes don’t break
    existing functionality.
6.  **Submit a pull request**: Push your changes to your fork and submit
    a pull request to the main repository.

#' Download topics from the World Bank API
#'
#' This function returns a tibble of supported topics for querying the World
#' Bank API. Topics represent the broad subject areas covered by the World
#' Bank's datasets.
#'
#' @param language A character string specifying the language code for the API
#'  response (default is "en" for English).
#'
#' @return A tibble with three columns:
#' \describe{
#'   \item{id}{The unique identifier for the topic.}
#'   \item{value}{The name of the topic (e.g., "Education", "Health").}
#'   \item{source_note}{A brief description or note about the topic.}
#' }
#'
#' @details This function provides a reference for the supported topics that can
#' be used to refine your queries when accessing the World Bank API. Topics
#' represent different areas of focus for data analysis.
#'
#' @source https://api.worldbank.org/v2/topics
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all available topics in English
#' wdi_get_topics()
#'
wdi_get_topics <- function(language = "en") {
  topics_raw <- perform_request("topics", language)

  if (!is.null(topics_raw)) {
    topics_processed <- as_tibble(topics_raw) |>
      mutate(id = as.integer(.data$id)) |>
      mutate(across(where(is.character), trimws)) |>
      select(topic_id = "id", topic_name = "value", topic_note = "sourceNote")

    topics_processed
  }
}

test_that("wdi_get_topics handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(result <- wdi_get_topics(language = "xx"))
  expect_null(result)
})

test_that("wdi_get_topics returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_topics()
  expect_s3_class(result, "tbl_df")
  expected_colnames <- c("topic_id", "topic_name", "topic_note")
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_topics trims whitespace in character columns", {
  mock_data <- tibble(
    id = c(" 1 ", " 2 "),
    value = c(" Education ", " Health "),
    sourceNote = c(" Covers educational data ", " Health indicators ")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_topics()
      expect_equal(result$topic_id, c(1, 2))
      expect_equal(result$topic_name, c("Education", "Health"))
      expect_equal(
        result$topic_note,
        c("Covers educational data", "Health indicators")
      )
    }
  )
})

test_that("wdi_get_topics converts id to integer", {
  mock_data <- tibble(
    id = c("1", "2"),
    value = c("Education", "Health"),
    sourceNote = c("Covers educational data", "Health indicators")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_topics()
      expect_type(result$topic_id, "integer")
      expect_equal(result$topic_id, c(1L, 2L))
    }
  )
})

test_that("wdi_get_topics handles empty data gracefully", {
  mock_data <- tibble(
    id = character(),
    value = character(),
    sourceNote = character()
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_topics()
      expect_equal(nrow(result), 0)
    }
  )
})

test_that("wdi_get_topics handles different language inputs", {
  skip_if_offline()

  result <- wdi_get_topics(language = "fr")
  expected_colnames <- c("topic_id", "topic_name", "topic_note")
  expect_true(all(expected_colnames %in% colnames(result)))
})

#' Download all available World Bank indicators
#'
#' This function retrieves a comprehensive list of all indicators supported by
#' the World Bank API. The indicators include metadata such as the indicator ID,
#' name, unit, source, and associated topics. The user can specify the language
#' of the API response and whether to include additional details.
#'
#' @param language A character string specifying the language for the request,
#'  see \link{wdi_get_languages}. Defaults to `"en"`.
#' @param per_page An integer specifying the number of results per page for the
#'  API. Defaults to 32,500. Must be a value between 1 and 32,500.
#'
#' @return A tibble containing the available indicators and their metadata:
#' \describe{
#'   \item{indicator_id}{A character string for the ID of the indicator (e.g.,
#'                       "NY.GDP.PCAP.KD").}
#'   \item{indicator_name}{A character string for the name of the indicator
#'                         (e.g., "GDP per capita, constant prices").}
#'   \item{source_id}{An integer identifying the data source providing the
#'                    indicator.}
#'   \item{source_name}{A character string describing the source of the
#'                      indicator data.}
#'   \item{source_note}{A character string providing additional notes about the
#'                      data source.}
#'   \item{source_organization}{A character string denoting the organization
#'                              responsible for the data source.}
#'   \item{topics}{A nested tibble containing (possibly multiple) topics
#'                 associated with the indicator, with two columns: an integer
#'                 \code{topic_id} and a character \code{topic_name}.}
#' }
#'
#' @details This function makes a request to the World Bank API to retrieve
#' metadata for all available indicators. It processes the response into a tidy
#' tibble format.
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' \donttest{
#' # Download all supported indicators in English
#' wdi_get_indicators()
#'
#' # Download all supported indicators in Spanish
#' wdi_get_indicators(language = "es")
#'}
#'
wdi_get_indicators <- function(language = "en", per_page = 32500L) {
  indicators_raw <- perform_request(
    "indicators",
    language = language,
    per_page = per_page
  )

  if (!is.null(indicators_raw)) {
    indicators_processed <- as_tibble(indicators_raw) |>
      rename(indicator_id = "id", indicator_name = "name") |>
      unnest_wider("source") |>
      rename(
        source_id = "id",
        source_name = "value",
        source_note = "sourceNote",
        source_organization = "sourceOrganization"
      ) |>
      select(-"unit") |>
      mutate(
        source_id = as.integer(.data$source_id),
        source_note = na_if(.data$source_note, ""),
        source_organization = na_if(.data$source_organization, "")
      )

    topics <- indicators_processed |>
      select("indicator_id", "topics") |>
      unnest_longer("topics") |>
      unnest_wider("topics") |>
      rename(topic_id = "id", topic_name = "value") |>
      mutate(
        topic_id = as.integer(.data$topic_id),
        topic_name = trimws(.data$topic_name)
      ) |>
      nest(topics = c("topic_id", "topic_name"))

    indicators_processed <- indicators_processed |>
      select(-"topics") |>
      left_join(topics, join_by("indicator_id"))

    indicators_processed
  }
}

test_that("wdi_get_indicators handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(result <- wdi_get_indicators(language = "xx"))
  expect_null(result)
})

test_that("wdi_get_indicators handles invalid per_page input", {
  expect_error(
    wdi_get_indicators(per_page = -1),
  )
  expect_error(
    wdi_get_indicators(per_page = 50000)
  )
  expect_error(
    wdi_get_indicators(per_page = "500")
  )
})

test_that("wdi_get_indicators returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_indicators()
  expect_s3_class(result, "tbl_df")

  expected_colnames <- c(
    "indicator_id",
    "indicator_name",
    "source_id",
    "source_name",
    "source_note",
    "source_organization",
    "topics"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_indicators handles type conversions and missing values", {
  mock_data <- tibble(
    id = "NY.GDP.PCAP.KD",
    name = "GDP per capita, constant prices",
    source = list(tibble(
      id = "2",
      value = "World Development Indicators",
      sourceNote = "Some note",
      sourceOrganization = "World Bank"
    )),
    unit = c(""),
    topics = list(list(list(id = "1", value = "Economy")))
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_indicators()
      expect_equal(result$indicator_id, "NY.GDP.PCAP.KD")
      expect_equal(result$indicator_name, "GDP per capita, constant prices")
      expect_equal(result$source_id, 2L)
      expect_equal(result$source_name, "World Development Indicators")
      expect_equal(result$source_note, "Some note")
      expect_equal(result$source_organization, "World Bank")
      expect_type(result$topics, "list")
      expect_equal(result$topics[[1]]$topic_id[1], 1L)
      expect_equal(result$topics[[1]]$topic_name[1], "Economy")
    }
  )
})

test_that("wdi_get_indicators handles different language inputs", {
  skip_if_offline()

  result <- wdi_get_indicators(language = "es")
  expected_colnames <- c(
    "indicator_id",
    "indicator_name",
    "source_id",
    "source_name",
    "source_note",
    "source_organization",
    "topics"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

#' Internal function to perform API requests to the World Bank API
#'
#' This function constructs and sends a request to the specified World Bank API
#' resource, with options to specify the language, the number of results per
#' page, a date range, and the data source. It also supports paginated requests
#' and progress tracking.
#'
#' @param resource A character string specifying the resource endpoint for the
#'  World Bank API (e.g., "incomeLevels", "lendingTypes").
#' @param language A character string specifying the language code for the API
#'  response. Defaults to NULL.
#' @param per_page An integer specifying the number of results per page for the
#'  API. Defaults to 10,000.
#' Must be a value between 1 and 32,500.
#' @param date A character string specifying the date range of the data to be
#'  retrieved (e.g., "2000:2020"). If NULL (default), no date filtering is
#'  applied.
#' @param source A character string specifying the data source for the API
#'  request. If NULL (default),
#' no specific source is selected.
#' @param progress A logical value indicating whether to display a progress bar
#'  for paginated requests. Defaults to FALSE.
#' @param base_url A character string specifying the base URL of the World Bank
#'  API (default is "https://api.worldbank.org/v2").
#' @param max_tries An integer specifying the number of attempts for the API
#'  request.
#'
#' @return A list containing the parsed JSON response from the API. If the
#'  response contains multiple pages, the results are combined into a single
#'  list. If the API returns an error, it provides the relevant error message
#'  and documentation.
#'
#' @details This internal helper function constructs the request URL using the
#' base URL, language, resource, and other optional parameters (such as date and
#' source), and performs the API request. For paginated results, it iterates
#' through all pages and consolidates the data into a single output. The
#' function handles API errors, providing detailed error messages when
#' available.
#'
#' @keywords internal
#'
perform_request <- function(
  resource,
  language = NULL,
  per_page = 10000L,
  date = NULL,
  most_recent_only = NULL,
  source = NULL,
  progress = FALSE,
  base_url = "https://api.worldbank.org/v2/",
  max_tries = 10L
) {
  validate_per_page(per_page)
  validate_max_tries(max_tries)

  req <- create_request(
    base_url,
    resource,
    language,
    per_page,
    date,
    most_recent_only,
    source
  ) |>
    req_retry(max_tries = max_tries)

  resp <- tryCatch(
    {
      resp <- req |>
        httr2::req_perform()
    },
    error = function(e) {
      cli::cli_alert_warning(
        paste(
          "Failed to retrieve data from the World Bank API",
          "for request {req_get_url(req)}.",
          "Error message: {conditionMessage(e)}"
        ),
        wrap = TRUE
      )
      invisible(NULL)
    }
  )

  if (!is.null(resp)) {
    if (is_request_error(resp)) {
      handle_request_error(resp)
      return(invisible(NULL))
    }

    body <- resp_body_json(resp, simplifyVector = TRUE)

    pages <- body[[1L]]$pages

    if (pages == 1L) {
      out <- body[[2L]]
    } else {
      resps <- req |>
        req_perform_iterative(
          next_req = iterate_with_offset("page"),
          max_reqs = pages,
          progress = progress
        )
      out <- resps |>
        purrr::map(function(x) {
          resp_body_json(x, simplifyVector = TRUE)[[2]]
        }) |>
        purrr::reduce(union)
    }

    out
  } else {
    resp
  }
}

#' @keywords internal
#' @noRd
validate_per_page <- function(per_page) {
  if (
    !is.numeric(per_page) ||
      per_page %% 1L != 0 ||
      per_page < 1L ||
      per_page > 32500L
  ) {
    cli::cli_abort("{.arg per_page} must be an integer between 1 and 32,500.")
  }
}

#' @keywords internal
#' @noRd
validate_max_tries <- function(max_tries) {
  if (!is.numeric(max_tries) || max_tries %% 1L != 0 || max_tries < 2L) {
    cli::cli_abort("{.arg max_tries} must be an integer larger than 1.")
  }
}

#' @keywords internal
#' @noRd
create_request <- function(
  base_url,
  resource,
  language,
  per_page,
  date,
  most_recent_only,
  source
) {
  request(base_url) |>
    req_url_path_append(language, resource) |>
    req_url_query(
      format = "json",
      per_page = per_page,
      date = date,
      source = source,
      mrv = if (isTRUE(most_recent_only)) 1L else NULL
    ) |>
    req_user_agent(
      "wbwdi R package (https://github.com/tidy-intelligence/r-wbwdi)"
    )
}

#' @keywords internal
#' @noRd
is_request_error <- function(resp) {
  status <- resp_status(resp)
  if (status >= 400L) {
    return(TRUE)
  }
  body <- resp_body_json(resp)
  if (length(body) == 1L && length(body[[1L]]$message) == 1L) {
    return(TRUE)
  }
  FALSE
}

#' @keywords internal
#' @noRd
check_for_body_error <- function(resp) {
  content_type <- httr2::resp_content_type(resp)
  if (identical(content_type, "application/json")) {
    body <- httr2::resp_body_json(resp)
    message_id <- body[[1]]$message[[1]]$id
    message_value <- body[[1]]$message[[1]]$value
    error_code <- paste("Error code:", message_id)
    docs <- paste0(
      "Read more at <https://datahelpdesk.worldbank.org/",
      "knowledgebase/articles/898620-api-error-codes>"
    )
    c(error_code, message_value, docs)
  }
}

#' @keywords internal
#' @noRd
handle_request_error <- function(resp) {
  error_body <- check_for_body_error(resp)
  cli::cli_alert_warning(paste(error_body, collapse = "\n"))
}

test_that("perform_request validates per_page parameter", {
  expect_error(
    perform_request("countries", per_page = 0),
    "per_page"
  )

  expect_error(
    perform_request("countries", per_page = 40000),
    "per_page"
  )
})

test_that("perform_request validates max_tries parameter", {
  expect_error(
    perform_request("countries", max_tries = 0),
    "max_tries"
  )

  expect_error(
    perform_request("countries", max_tries = -1),
    "max_tries"
  )
})

test_that("perform_request constructs correct URL with all parameters", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries",
    language = "en",
    per_page = 50,
    date = "2020:2021",
    source = 2
  )

  expect_type(result, "list")
})

test_that("perform_request handles single page response", {
  skip_if_offline()

  result <- perform_request(
    resource = "incomeLevels",
    per_page = 100
  )

  expect_type(result, "list")
  expect_true(length(result) > 0)
})

test_that("perform_request handles paginated responses", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries",
    per_page = 10,
    progress = FALSE
  )

  expect_type(result, "list")
  expect_true(nrow(result) > 10)
})

test_that("perform_request works with progress bar enabled", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries",
    per_page = 10,
    progress = TRUE
  )

  expect_type(result, "list")
})

test_that("perform_request handles language parameter", {
  skip_if_offline()

  result_en <- perform_request(
    resource = "countries",
    language = "en",
    per_page = 10
  )

  result_es <- perform_request(
    resource = "countries",
    language = "es",
    per_page = 10
  )

  expect_type(result_en, "list")
  expect_type(result_es, "list")
})

test_that("perform_request handles date parameter", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries/USA/indicators/SP.POP.TOTL",
    date = "2015:2020",
    per_page = 100
  )

  expect_type(result, "list")
})

test_that("perform_request handles most_recent_only parameter", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries/USA/indicators/SP.POP.TOTL",
    most_recent_only = 1,
    per_page = 100
  )

  expect_type(result, "list")
})

test_that("perform_request handles source parameter", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries/USA/indicators/SP.POP.TOTL",
    source = 2,
    per_page = 100
  )

  expect_type(result, "list")
})

test_that("perform_request handles invalid resource gracefully", {
  skip_if_offline()

  expect_message(
    result <- perform_request(
      resource = "nonexistent_resource_xyz123",
      per_page = 10
    ),
    "Failed to retrieve data"
  )

  expect_null(result)
})

test_that("perform_request handles network errors gracefully", {
  expect_message(
    result <- perform_request(
      resource = "countries",
      base_url = "https://invalid.worldbank.org.invalid/v2/",
      max_tries = 2
    ),
    "Failed to retrieve data"
  )

  expect_null(result)
})

test_that("perform_request uses custom base_url", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries",
    base_url = "https://api.worldbank.org/v2/",
    per_page = 10
  )

  expect_type(result, "list")
})

test_that("perform_request returns NULL on connection failure", {
  expect_message(
    result <- perform_request(
      resource = "countries",
      base_url = "https://localhost:9999/",
      max_tries = 2
    )
  )

  expect_null(result)
})

test_that("perform_request combines multiple pages correctly", {
  skip_if_offline()

  result_small <- perform_request(
    resource = "countries",
    per_page = 5
  )

  result_large <- perform_request(
    resource = "countries",
    per_page = 10000
  )

  expect_equal(length(result_small), length(result_large))
})

test_that("perform_request handles NULL optional parameters", {
  skip_if_offline()

  result <- perform_request(
    resource = "countries",
    language = NULL,
    date = NULL,
    most_recent_only = NULL,
    source = NULL
  )

  expect_type(result, "list")
})

test_that("perform_request max_tries parameter works", {
  start_time <- Sys.time()

  expect_message(
    result <- perform_request(
      resource = "countries",
      base_url = "https://localhost:9999/",
      max_tries = 2
    )
  )

  end_time <- Sys.time()

  expect_null(result)
  expect_true(as.numeric(end_time - start_time, units = "secs") < 10)
})

test_that("is_request_error identifies error responses correctly", {
  mock_resp <- structure(list(status_code = 404), class = "httr2_response")
  expect_true(is_request_error(mock_resp))
})

test_that("returns FALSE when body$message has length > 1", {
  dummy <- structure(list(), class = "resp")
  local_mocked_bindings(
    resp_status = function(resp) 200L,
    resp_body_json = function(resp) list(list(message = c("a", "b")))
  )
  expect_false(is_request_error(dummy))
})


test_that("perform_request handles API errors gracefully", {
  skip_if_offline()

  expect_message(perform_request("nonexistent"), "HTTP 404 Not Found.")
})

test_that("perform_request returns NULL w/ warning when body contains error", {
  mock_resp <- structure(list(status_code = 200L), class = "httr2_response")
  mock_req <- structure(list(), class = "httr2_request")

  with_mocked_bindings(
    req_retry = function(req, ...) mock_req,
    req_perform = function(req, ...) mock_resp,
    req_get_url = function(req) "https://api.worldbank.org/v2/test",
    is_request_error = function(resp) TRUE,
    check_for_body_error = function(resp) c("Error code: 120", "Invalid value"),
    {
      expect_message(
        result <- perform_request("test", max_tries = 2L)
      )
      expect_null(result)
    }
  )
})

test_that("perform_request returns NULL with warning on pagination error", {
  mock_resp <- structure(list(status_code = 200L), class = "httr2_response")
  mock_req <- structure(list(), class = "httr2_request")
  mock_body <- list(
    list(pages = 2L),
    list(list(value = 1))
  )

  with_mocked_bindings(
    req_retry = function(req, ...) mock_req,
    req_perform = function(req, ...) mock_resp,
    req_get_url = function(req) "https://api.worldbank.org/v2/test",
    is_request_error = function(resp) FALSE,
    resp_body_json = function(resp, ...) mock_body,
    req_perform_iterative = function(...) stop("HTTP 502 Bad Gateway"),
    {
      expect_message(
        result <- perform_request("test", max_tries = 2L),
        "Failed to retrieve data"
      )
      expect_null(result)
    }
  )
})

test_that("perform_request warns and returns NULL when req_perform fails", {
  # Stub req_perform to simulate a network/transport failure
  local_mocked_bindings(
    req_perform = function(req, ...) {
      stop("Could not resolve host: api.worldbank.org")
    },
    .package = "httr2"
  )

  expect_message(
    result <- perform_request(resource = "country", max_tries = 2L),
    regexp = "Failed to retrieve data from the World Bank API"
  )

  expect_null(result)
})

test_that("perform_request warning includes request URL and error message", {
  local_mocked_bindings(
    req_perform = function(req, ...) {
      stop("Timeout was reached")
    },
    .package = "httr2"
  )

  expect_message(
    perform_request(resource = "country", max_tries = 2L),
    regexp = "Timeout was reached"
  )

  expect_message(
    perform_request(resource = "country", max_tries = 2L),
    regexp = "api\\.worldbank\\.org"
  )
})

#' Download World Bank indicator data for specific entities and time periods
#'
#' This function retrieves indicator data from the World Bank API for a
#' specified set of entities and indicators. The user can specify one or more
#' indicators, a date range, and other options to tailor the request. The data
#' is processed and returned in a tidy format, including country, indicator,
#' date, and value fields.
#'
#' @param entities A character vector of ISO 2 or ISO 3-country codes, or
#'  `"all"` to retrieve data for all entities.
#' @param indicators A character vector specifying one or more World Bank
#'  indicators to download (e.g., c("NY.GDP.PCAP.KD", "SP.POP.TOTL")).
#' @param start_year Optional integer. The starting date for the data as a year.
#' @param end_year Optional integer. The ending date for the data as a year.
#' @param most_recent_only A logical value indicating whether to download only
#'  the most recent value. In case of `TRUE`, it overrides `start_year` and
#'  `end_year`. Defaults to `FALSE`.
#' @param frequency A character string specifying the frequency of the data
#'  ("annual", "quarter", "month"). Defaults to "annual".
#' @param language A character string specifying the language for the request,
#'  see \link{wdi_get_languages}. Defaults to `"en"`.
#' @param per_page An integer specifying the number of results per page for the
#'  API. Defaults to 10,000.
#' @param progress A logical value indicating whether to show progress messages
#'  during the data download and parsing. Defaults to `TRUE`.
#' @param source An integer value specifying the data source, see
#'  \link{wdi_get_sources}.
#' @param format A character value specifying whether the data is returned in
#'  `"long"` or `"wide"` format. Defaults to `"long"`.
#'
#' @return A tibble with the following columns:
#' \describe{
#'   \item{entity_id}{The ISO 3-country code of the country or aggregate for
#'                    which the data was retrieved.}
#'   \item{indicator_id}{The ID of the indicator (e.g., "NY.GDP.PCAP.KD").}
#'   \item{year}{The year of the indicator data as an integer.}
#'   \item{quarter}{Optional. The quarter of the indicator data as integer.}
#'   \item{month}{Optional. The month of the indicator data as integer.}
#'   \item{value}{The value of the indicator for the given country and date.}
#' }
#'
#' @details This function constructs a request URL for the World Bank API,
#' retrieves the relevant data for the given entities and indicators, and
#' processes the response into a tidy format. The user can optionally specify a
#' date range, and the function will handle requests for multiple pages if
#' necessary. If the `progress` parameter is `TRUE`, messages will be displayed
#' during the request and parsing process.
#'
#' The function supports downloading multiple indicators by sending individual
#' API requests for each indicator and then combining the results into a single
#' tidy data frame.
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' \donttest{
#' # Download single indicator for multiple entities
#' wdi_get(c("USA", "CAN", "GBR"), "NY.GDP.PCAP.KD")
#'
#' # Download single indicator for a specific time frame
#' wdi_get(c("USA", "CAN", "GBR"), "DPANUSSPB",
#'         start_year = 2012, end_year = 2013)
#'
#' # Download single indicator for monthly frequency
#' wdi_get("AUT", "DPANUSSPB",
#'         start_year = 2012, end_year = 2015, frequency = "month")
#'
#' # Download single indicator for quarterly frequency
#' wdi_get("NGA", "DT.DOD.DECT.CD.TL.US",
#'         start_year = 2012, end_year = 2015, frequency = "quarter")
#'
#' # Download single indicator for all entities and disable progress bar
#' wdi_get("all", "NY.GDP.PCAP.KD", progress = FALSE)
#'
#' # Download multiple indicators for multiple entities
#' wdi_get(c("USA", "CAN", "GBR"), c("NY.GDP.PCAP.KD", "SP.POP.TOTL"))
#'
#' # Download indicators for different sources
#' wdi_get("DEU", "NY.GDP.PCAP.KD", source = 2)
#' wdi_get("DEU", "SG.LAW.INDX", source = 14)
#'
#' # Download indicators in wide format
#' wdi_get(c("USA", "CAN", "GBR"), c("NY.GDP.PCAP.KD"),
#'         format = "wide")
#' wdi_get(c("USA", "CAN", "GBR"), c("NY.GDP.PCAP.KD", "SP.POP.TOTL"),
#'         format = "wide")
#'
#' # Download most recent value only
#' wdi_get("USA", "SP.POP.TOTL", most_recent_only = TRUE)
#' }
wdi_get <- function(
  entities,
  indicators,
  start_year = NULL,
  end_year = NULL,
  most_recent_only = FALSE,
  frequency = "annual",
  language = "en",
  per_page = 10000L,
  progress = TRUE,
  source = NULL,
  format = "long"
) {
  validate_most_recent_only(most_recent_only)
  validate_frequency(frequency)
  validate_progress(progress)
  validate_source(source)
  validate_format(format)

  if (!most_recent_only) {
    if (frequency == "annual") {
      start_year <- as.character(start_year)
      end_year <- as.character(end_year)
    }
    if (frequency == "quarter") {
      start_year <- paste0(start_year, "Q1")
      end_year <- paste0(end_year, "Q4")
    }
    if (frequency == "month") {
      start_year <- paste0(start_year, "M01")
      end_year <- paste0(end_year, "M12")
    }
  }

  indicators_processed <- indicators |>
    map_df(
      ~ get_indicator(
        .,
        entities,
        start_year,
        end_year,
        most_recent_only,
        language,
        per_page,
        progress,
        source
      )
    )

  if (nrow(indicators_processed) > 0) {
    if (format == "wide") {
      indicators_processed <- indicators_processed |>
        tidyr::pivot_wider(names_from = "indicator_id", values_from = "value")
    }

    # ISO3 field is not always populated and sometimes the id already is ISO3
    if (
      nrow(indicators_processed) > 0 &&
        nchar(indicators_processed$entity_id[1]) == 2
    ) {
      entities <- wdi_get_entities()

      if (!is.null(entities)) {
        indicators_processed <- indicators_processed |>
          rename(entity_iso2code = "entity_id") |>
          left_join(
            entities |>
              select("entity_id", "entity_iso2code"),
            join_by("entity_iso2code")
          ) |>
          select(-"entity_iso2code")
      }
    }
  }

  indicators_processed <- indicators_processed |>
    select("entity_id", everything())

  indicators_processed
}

#' @keywords internal
#' @noRd
validate_most_recent_only <- function(most_recent_only) {
  if (!is.logical(most_recent_only)) {
    cli::cli_abort("{.arg most_recent_only} must be either TRUE or FALSE.")
  }
}

#' @keywords internal
#' @noRd
validate_frequency <- function(frequency) {
  valid_frequencies <- c("annual", "quarter", "month")
  if (!frequency %in% valid_frequencies) {
    cli::cli_abort(
      "{.arg frequency} must be either 'annual', 'quarter', or 'month'."
    )
  }
}

#' @keywords internal
#' @noRd
validate_progress <- function(progress) {
  if (!is.logical(progress)) {
    cli::cli_abort("{.arg progress} must be either TRUE or FALSE.")
  }
}

#' @keywords internal
#' @noRd
validate_source <- function(source) {
  if (!is.null(source)) {
    supported_sources <- wdi_get_sources()
    if (!is.null(supported_sources)) {
      if (!source %in% supported_sources$source_id) {
        cli::cli_abort(
          paste(
            "{.arg source} is not supported.",
            "Please call {.fun wdi_get_sources()}."
          )
        )
      }
    } else {
      cli::cli_alert_warning(
        "The provided {.arg source} could not be verified. "
      )
    }
  }
}

#' @keywords internal
#' @noRd
validate_format <- function(format) {
  if (!is.character(format) || !format %in% c("long", "wide")) {
    cli::cli_abort("{.arg format} must be either 'long' or 'wide'.")
  }
}

#' @keywords internal
#' @noRd
create_date <- function(start_year, end_year) {
  if (!is.null(start_year) && !is.null(end_year)) {
    paste0(start_year, ":", end_year)
  }
}

#' @keywords internal
#' @noRd
get_indicator <- function(
  indicator,
  entities,
  start_year,
  end_year,
  most_recent_only,
  language,
  per_page,
  progress,
  source
) {
  if (progress) {
    progress_req <- paste0("Sending requests for indicator ", indicator)
  } else {
    progress_req <- FALSE
  }

  date <- create_date(start_year, end_year)

  resource <- paste0(
    "country/",
    paste(entities, collapse = ";"),
    "/indicator/",
    indicator
  )

  indicator_raw <- perform_request(
    resource,
    language,
    per_page,
    date,
    most_recent_only,
    source,
    progress_req
  )

  if (!is.null(indicator_raw)) {
    indicator_parsed <- as_tibble(indicator_raw) |>
      rename(.value = "value") |>
      unnest_wider("indicator") |>
      rename(indicator_id = "id") |>
      select(-"value") |>
      unnest_wider("country") |>
      rename(entity_id = "id") |>
      select(-"value") |>
      select("indicator_id", "entity_id", "date", ".value") |>
      rename(value = ".value") |>
      mutate(value = as.numeric(.data$value))

    if (grepl("Q", indicator_parsed$date[1], fixed = TRUE)) {
      indicator <- indicator_parsed |>
        mutate(
          year = as.integer(substr(.data$date, 1, 4)),
          quarter = as.integer(substr(.data$date, 6, 6))
        ) |>
        select(-"date") |>
        arrange(.data$year, .data$quarter)
    } else if (grepl("M", indicator_parsed$date[1], fixed = TRUE)) {
      indicator <- indicator_parsed |>
        mutate(
          year = as.integer(substr(.data$date, 1, 4)),
          month = as.integer(substr(.data$date, 6, 7))
        ) |>
        select(-"date") |>
        arrange(.data$year, .data$month)
    } else {
      indicator <- indicator_parsed |>
        mutate(year = as.integer(.data$date)) |>
        select(-"date") |>
        arrange(.data$year)
    }

    indicator <- indicator |>
      relocate("value", .after = last_col())

    indicator
  } else {
    tibble(
      entity_id = character(),
      indicator_id = character(),
      year = integer(),
      value = numeric()
    )
  }
}

test_that("wdi_get handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(
    result <- wdi_get(
      entities = "USA",
      indicators = "NY.GDP.MKTP.CD",
      language = "xx",
      progress = FALSE
    )
  )
  expect_equal(nrow(result), 0)
})

test_that("wdi_get checks invalid parameter values", {
  expect_error(
    wdi_get("USA", "NY.GDP.PCAP.KD", frequency = "weekly"),
    "must be either 'annual', 'quarter', or 'month'"
  )
  expect_error(
    wdi_get("USA", "NY.GDP.PCAP.KD", format = "tall"),
    "must be either 'long' or 'wide'"
  )
})

test_that("wdi_get handles invalid per_page input", {
  expect_error(
    wdi_get(
      entities = "USA",
      indicators = "NY.GDP.MKTP.CD",
      per_page = -1
    )
  )
  expect_error(
    wdi_get(
      entities = "USA",
      indicators = "NY.GDP.MKTP.CD",
      per_page = "1000"
    )
  )
})

test_that("wdi_get handels invalid most_recent_only input", {
  expect_error(
    wdi_get(
      entities = "USA",
      indicators = "NY.GDP.MKTP.CD",
      most_recent_only = "yes"
    )
  )
})

test_that("wdi_get handels invalid progress input", {
  expect_error(
    wdi_get(
      entities = "USA",
      indicators = "NY.GDP.MKTP.CD",
      progress = "yes"
    )
  )
})

test_that("wdi_get creates valid output structure for single indicator", {
  skip_if_offline()

  result <- wdi_get(
    entities = "USA",
    indicators = "NY.GDP.MKTP.CD",
    start_year = 2010,
    end_year = 2020,
    language = "en",
    per_page = 10,
    progress = FALSE
  )
  expected_names <- c("indicator_id", "entity_id", "year", "value")
  expect_true(is.data.frame(result))
  expect_true(all(expected_names %in% names(result)))
  expect_equal(nrow(result), 11)
})

test_that("wdi_get_creates valid output structure for multiple indicators", {
  skip_if_offline()

  result <- wdi_get(
    entities = "USA",
    indicators = c("NY.GDP.MKTP.CD", "SP.POP.TOTL"),
    start_year = 2010,
    end_year = 2020,
    language = "en",
    per_page = 10,
    progress = FALSE
  )
  expected_names <- c("indicator_id", "entity_id", "year", "value")
  expect_true(is.data.frame(result))
  expect_true(all(expected_names %in% names(result)))
  expect_true(any(result$indicator_id == "NY.GDP.MKTP.CD"))
  expect_true(any(result$indicator_id == "SP.POP.TOTL"))
})

test_that("wdi_get returns a tibble", {
  skip_if_offline()

  result <- wdi_get("USA", "NY.GDP.PCAP.KD")
  expect_s3_class(result, "tbl_df")
})

test_that("wdi_get handles single indicator, entity & default parameters", {
  skip_if_offline()

  result <- wdi_get("USA", "NY.GDP.PCAP.KD")
  expected_colnames <- c("indicator_id", "entity_id", "year", "value")
  expect_true(all(expected_colnames %in% colnames(result)))
  expect_equal(unique(result$entity_id), "USA")
  expect_equal(unique(result$indicator_id), "NY.GDP.PCAP.KD")
})

test_that("wdi_get handles multiple indicators and multiple entities", {
  skip_if_offline()

  result <- wdi_get(c("USA", "CAN"), c("NY.GDP.PCAP.KD", "SP.POP.TOTL"))
  expected_colnames <- c("indicator_id", "entity_id", "year", "value")
  expect_true(all(expected_colnames %in% colnames(result)))
  expect_true(all(result$entity_id %in% c("USA", "CAN")))
  expect_true(all(result$indicator_id %in% c("NY.GDP.PCAP.KD", "SP.POP.TOTL")))
})

test_that("wdi_get handles different date ranges and frequencies", {
  skip_if_offline()

  result_annual <- wdi_get(
    "USA",
    "NY.GDP.PCAP.KD",
    start_year = 2010,
    end_year = 2015,
    frequency = "annual"
  )
  result_quarter <- wdi_get(
    "NGA",
    "DT.DOD.DECT.CD.TL.US",
    start_year = 2010,
    end_year = 2015,
    frequency = "quarter"
  )
  result_month <- wdi_get(
    "USA",
    "DPANUSSPB",
    start_year = 2010,
    end_year = 2015,
    frequency = "month"
  )

  expect_equal(range(result_annual$year), c(2010, 2015))
  expect_true("quarter" %in% colnames(result_quarter))
  expect_true("month" %in% colnames(result_month))
})

test_that("wdi_get handles format parameter (long and wide)", {
  skip_if_offline()

  result_long <- wdi_get("USA", "NY.GDP.PCAP.KD", format = "long")
  result_wide <- wdi_get("USA", "NY.GDP.PCAP.KD", format = "wide")

  expect_true("indicator_id" %in% colnames(result_long))
  expect_false("indicator_id" %in% colnames(result_wide))
})

test_that("wdi_get handles empty data gracefully", {
  mock_data <- data.frame(
    indicator = I(data.frame(
      id = character(),
      value = character()
    )),
    country = I(data.frame(
      id = character(),
      value = character()
    )),
    countryiso3code = character(),
    date = character(),
    value = numeric(),
    unit = character(),
    obs_status = character(),
    decimal = integer()
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get("USA", "NY.GDP.PCAP.KD")
      expect_equal(nrow(result), 0)
    }
  )
})

test_that("wdi_get handles NULL from wdi_get_entities during ISO2 enrichment", {
  mock_data <- data.frame(
    indicator = I(data.frame(id = "NY.GDP.MKTP.CD", value = "GDP")),
    country = I(data.frame(id = "US", value = "United States")),
    countryiso3code = "USA",
    date = "2020",
    value = 21000000000000,
    unit = "",
    obs_status = "",
    decimal = 0L
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    wdi_get_entities = function(...) NULL,
    {
      result <- wdi_get("US", "NY.GDP.MKTP.CD", progress = FALSE)
      expect_s3_class(result, "tbl_df")
      expect_true(nrow(result) > 0)
    }
  )
})

test_that("create_date constructs date range correctly", {
  expect_equal(create_date(2010, 2015), "2010:2015")
  expect_null(create_date(NULL, NULL))
})

test_that("validate_frequency checks valid frequencies", {
  expect_error(
    validate_frequency("weekly"),
    "must be either 'annual', 'quarter', or 'month'"
  )
  expect_silent(validate_frequency("annual"))
  expect_silent(validate_frequency("quarter"))
  expect_silent(validate_frequency("month"))
})

test_that("validate_progress checks if progress is logical", {
  expect_error(validate_progress("yes"), "must be either TRUE or FALSE")
  expect_silent(validate_progress(TRUE))
  expect_silent(validate_progress(FALSE))
})

test_that("validate_source checks if source is valid", {
  with_mocked_bindings(
    wdi_get_sources = function(...) tibble(source_id = c(1, 2, 3)),
    {
      expect_error(validate_source(4), "is not supported")
      expect_silent(validate_source(1))
      expect_silent(validate_source(NULL))
    }
  )
})

test_that("validate_format checks valid formats", {
  expect_error(validate_format("tall"), "must be either 'long' or 'wide'")
  expect_silent(validate_format("long"))
  expect_silent(validate_format("wide"))
})

test_that("prints a CLI warning when sources cannot be verified (NULL)", {
  with_mocked_bindings(
    wdi_get_sources = function() NULL,
    {
      expect_message(
        validate_source("anything"),
        regexp = "could not be verified"
      )
    }
  )
})

test_that("get_indicator returns empty tibble when indicator_raw is NULL", {
  result <- with_mocked_bindings(
    get_indicator(
      indicator = "NY.GDP.MKTP.CD",
      entities = "USA",
      start_year = 2020,
      end_year = 2021,
      most_recent_only = FALSE,
      language = "en",
      per_page = 100,
      progress = FALSE,
      source = NULL
    ),
    perform_request = function(...) NULL
  )

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 0)
  expect_equal(ncol(result), 4)
  expect_named(result, c("entity_id", "indicator_id", "year", "value"))

  expect_type(result$entity_id, "character")
  expect_type(result$indicator_id, "character")
  expect_type(result$year, "integer")
  expect_type(result$value, "double")
})

#' Download data sources from the World Bank API
#'
#' This function returns a tibble of supported data sources for querying the
#' World Bank API. The data sources include various databases and datasets
#' provided by the World Bank.
#'
#' @param language A character string specifying the language code for the API
#'  response (default is "en" for English).
#'
#' @return A tibble with six columns:
#' \describe{
#'   \item{source_id}{An integer identifier for the data source.}
#'   \item{source_code}{A character string for the source code.}
#'   \item{source_name}{The name of the data source (e.g., "World Development
#'                      Indicators").}
#'   \item{update_date}{The last update date of the data source.}
#'   \item{is_data_available}{A boolean indicating whether data is available.}
#'   \item{is_metadata_available}{A boolean indicating whether metadata is
#'                                available.}
#'   \item{concepts}{The number of concepts defined for the data source.}
#' }
#'
#' @details This function provides a reference for the supported data sources
#' and their metadata when querying the World Bank API. The columns
#' `is_data_available` and `is_metadata_available` are logical values
#' derived from the API response, where "Y" indicates availability.
#'
#' @source https://api.worldbank.org/v2/sources
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all available data sources in English
#' wdi_get_sources()
#'
wdi_get_sources <- function(language = "en") {
  sources_raw <- perform_request("sources", language)

  # url and description are always empty, hence omitted
  if (!is.null(sources_raw)) {
    sources_processed <- as_tibble(sources_raw) |>
      select(
        source_id = "id",
        source_code = "code",
        source_name = "name",
        update_date = "lastupdated",
        is_data_available = "dataavailability",
        is_metadata_available = "metadataavailability",
        concepts = "concepts"
      ) |>
      mutate(
        across(c("is_data_available", "is_metadata_available"), ~ . == "Y"),
        update_date = as.Date(.data$update_date),
        across(c("source_id", "concepts"), as.integer),
        across(where(is.character), trimws)
      )

    sources_processed
  }
}

test_that("wdi_get_sources handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(result <- wdi_get_sources(language = "xx"))
  expect_null(result)
})

test_that("wdi_get_sources trims whitespace in character columns", {
  mock_data <- tibble(
    id = c(" 1 ", " 2 "),
    code = c(" WDI ", " GFDD "),
    name = c(
      " World Development Indicators ",
      " Global Financial Development "
    ),
    lastupdated = c(" 2024-10-01 ", " 2024-11-01 "),
    dataavailability = c("Y", "N"),
    metadataavailability = c("Y", "Y"),
    concepts = c(" 100 ", " 200 ")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_sources()
      expect_equal(result$source_id, c(1L, 2L))
      expect_equal(result$source_code, c("WDI", "GFDD"))
      expect_equal(
        result$source_name,
        c("World Development Indicators", "Global Financial Development")
      )
      expect_equal(result$update_date, as.Date(c("2024-10-01", "2024-11-01")))
      expect_equal(result$is_data_available, c(TRUE, FALSE))
      expect_equal(result$is_metadata_available, c(TRUE, TRUE))
      expect_equal(result$concepts, c(100L, 200L))
    }
  )
})

test_that("wdi_get_sources converts data types correctly", {
  mock_data <- tibble(
    id = c("1", "2"),
    code = c("WDI", "GFDD"),
    name = c("World Development Indicators", "Global Financial Development"),
    lastupdated = c("2024-10-01", "2024-11-01"),
    dataavailability = c("Y", "N"),
    metadataavailability = c("Y", "Y"),
    concepts = c("100", "200")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_sources()
      expect_type(result$source_id, "integer")
      expect_type(result$update_date, "double")
      expect_type(result$is_data_available, "logical")
      expect_type(result$is_metadata_available, "logical")
      expect_type(result$concepts, "integer")
    }
  )
})

test_that("wdi_get_sources handles empty data gracefully", {
  mock_data <- tibble(
    id = character(),
    code = character(),
    name = character(),
    lastupdated = character(),
    dataavailability = character(),
    metadataavailability = character(),
    concepts = character()
  )

  with_mocked_bindings(
    perform_request = function(endpoint, language) mock_data,
    {
      result <- wdi_get_sources()
      expect_equal(nrow(result), 0)
    }
  )
})

#' Search for Keywords in Data Frame Columns
#'
#' This function searches for specified keywords across columns in a data frame
#' and returns only the rows where at least one keyword appears in any of the
#' specified columns. It supports nested columns (lists within columns) and
#' provides the option to limit the search to specific columns.
#'
#' @param data A data frame to search. It can include nested columns (lists
#'  within columns).
#' @param keywords A character vector of keywords to search for within the
#'  specified columns. The search is case-insensitive.
#' @param columns Optional. A character vector of column names to limit the
#'  search to specific columns. If `NULL`, all columns in `data` will be
#'  searched.
#'
#' @return A data frame containing only the rows where at least one keyword is
#'         found in the specified columns. The returned data frame has the same
#'         structure as `data`.
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' \donttest{
#' # Download indicators
#' indicators <- wdi_get_indicators()
#'
#' # Search for keywords "inequality" or "gender" across all columns
#' wdi_search(
#'   indicators,
#'   keywords = c("inequality", "gender")
#' )
#'
#' # Search for keywords only within the "indicator_name" column
#' wdi_search(
#'   indicators,
#'   keywords = c("inequality", "gender"),
#'   columns = c("indicator_name")
#' )
#'}
wdi_search <- function(data, keywords, columns = NULL) {
  if (is.null(columns)) {
    columns_to_search <- colnames(data)
  } else {
    columns_to_search <- intersect(columns, colnames(data))
  }

  row_matches <- purrr::map_lgl(seq_len(nrow(data)), function(i) {
    any(purrr::map_lgl(
      data[columns_to_search],
      ~ {
        col_value <- .x[[i]]
        if (is.list(col_value) && length(col_value) > 0) {
          any(
            purrr::map_lgl(
              unlist(col_value, recursive = TRUE, use.names = FALSE),
              ~ contains_keyword(., keywords) %||% FALSE
            )
          )
        } else if (!is.null(col_value) && length(col_value) > 0) {
          contains_keyword(col_value, keywords) %||% FALSE
        } else {
          FALSE
        }
      }
    ))
  })

  data[row_matches, , drop = FALSE]
}

#' @keywords internal
#' @noRd
contains_keyword <- function(value, keywords) {
  value <- as.character(value)
  any(purrr::map_lgl(keywords, ~ grepl(.x, value, ignore.case = TRUE)))
}

test_that("wdi_search returns rows with specified keywords in any column", {
  data <- tibble::tibble(
    id = 1:4,
    text = c("apple pie", NA, "", "banana split"),
    description = c("A delicious apple dessert", "Tropical fruit mix", "", NA),
    nested = list(list("apple", "orange"), list(), list(), list("banana"))
  )

  # Test 1: Search for a keyword in all columns
  result <- wdi_search(data, keywords = c("apple"))
  expect_equal(nrow(result), 1)

  # Test 2: Search for multiple keywords across all columns
  result <- wdi_search(data, keywords = c("apple", "berry"))
  expect_equal(nrow(result), 1)

  # Test 3: Search for keywords only in a specific column
  result <- wdi_search(data, keywords = c("banana"), columns = "text")
  expect_equal(nrow(result), 1)

  # Test 4: Search for keywords in multiple specified columns
  result <- wdi_search(
    data,
    keywords = c("pie"),
    columns = c("text", "description")
  )
  expect_equal(nrow(result), 1)

  # Test 5: Case-insensitive search
  result <- wdi_search(data, keywords = c("APPLE"))
  expect_equal(nrow(result), 1)

  # Test 6: Keyword not present in any row
  result <- wdi_search(data, keywords = c("mango"))
  expect_equal(nrow(result), 0)

  # Test 7: Nested column search
  result <- wdi_search(data, keywords = c("orange"))
  expect_equal(nrow(result), 1)

  # Test 8: All columns searched by default if columns parameter is NULL
  result <- wdi_search(data, keywords = c("banana"))
  expect_equal(nrow(result), 1)

  # Test 9: Empty keywords vector
  result <- wdi_search(data, keywords = character(0))
  expect_equal(nrow(result), 0)

  # Test 10: No relevant keyword
  result <- wdi_search(data, keywords = c("xxx"))
  expect_equal(nrow(result), 0)

  # Test 11: All rows match when keyword is common
  result <- wdi_search(data, keywords = c("a"))
  expect_equal(nrow(result), 3)

  # Test 12: Search with NULL columns explicitly
  result <- wdi_search(data, keywords = c("tropical"), columns = NULL)
  expect_equal(nrow(result), 1)
  expect_equal(result$id, 2)

  # Test 13: Search for a keyword that doesn't exist in empty/null cells
  result <- wdi_search(data, keywords = c("mango"))
  expect_equal(nrow(result), 0)

  # Test 14: Search in columns with NULL and empty lists
  result <- wdi_search(data, keywords = c("banana"))
  expect_equal(nrow(result), 1)
  expect_equal(result$id, 4)
})

#' Download income levels from the World Bank API
#'
#' This function returns a tibble of supported income levels for querying the
#' World Bank API. The income levels categorize countries based on their gross
#' national income per capita.
#'
#' @param language A character string specifying the language code for the API
#'  response (default is "en" for English).
#'
#' @return A tibble with columns that typically include:
#' \describe{
#'   \item{income_level_id}{An integer identifier for the income level.}
#'   \item{income_level_iso2code}{A character string representing the ISO2 code
#'                                for the income level.}
#'   \item{income_level_name}{The description of the income level (e.g.,
#'                            "Low income", "High income").}
#' }
#'
#' @details This function provides a reference for the supported income levels,
#' which categorize countries according to their income group as defined by the
#' World Bank. The language parameter allows the results to be returned in
#' different languages as supported by the API.
#'
#' @source https://api.worldbank.org/v2/incomeLevels
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all income levels in English
#' wdi_get_income_levels()
#'
wdi_get_income_levels <- function(language = "en") {
  income_levels_raw <- perform_request("incomeLevels", language)

  if (!is.null(income_levels_raw)) {
    income_levels_processed <- as_tibble(income_levels_raw) |>
      rename(
        income_level_id = "id",
        income_level_iso2code = "iso2code",
        income_level_name = "value"
      ) |>
      mutate(across(where(is.character), trimws))

    income_levels_processed
  }
}

test_that("wdi_get_income_levels returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_income_levels()
  expect_s3_class(result, "tbl_df")
  expected_colnames <- c(
    "income_level_id",
    "income_level_iso2code",
    "income_level_name"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_income_levels trims whitespace in character columns", {
  mock_data <- tibble(
    id = c(" LIC ", " HIC "),
    iso2code = c(" 1 ", " 2 "),
    value = c(" Low income ", " High income ")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_income_levels()
      expect_equal(result$income_level_id, c("LIC", "HIC"))
      expect_equal(result$income_level_iso2code, c("1", "2"))
      expect_equal(result$income_level_name, c("Low income", "High income"))
    }
  )
})

test_that("wdi_get_income_levels handles different language inputs", {
  skip_if_offline()

  result <- wdi_get_income_levels(language = "es")
  expected_colnames <- c(
    "income_level_id",
    "income_level_iso2code",
    "income_level_name"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_income_levels handles empty data gracefully", {
  mock_data <- tibble(
    id = character(),
    iso2code = character(),
    value = character()
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_income_levels()
      expect_equal(nrow(result), 0)
    }
  )
})

#' Download regions from the World Bank API
#'
#' This function returns a tibble of supported regions for querying the World
#' Bank API.The regions include various geographic areas covered by the World
#' Bank's datasets.
#'
#' @param language A character string specifying the language code for the API
#'  response (default is "en" for English).
#'
#' @return A tibble with the following columns:
#' \describe{
#'   \item{region_id}{An integer identifier for each region.}
#'   \item{region_code}{A character string representing the region code.}
#'   \item{region_iso2code}{A character string representing the ISO2 code for
#'                          the region.}
#'   \item{region_name}{A character string representing the name of the region,
#'                      in the specified language.}
#' }
#'
#' @details This function provides a reference for the supported regions, which
#'  are important for refining queries related to geographic data in the World
#'  Bank's datasets. The `region_id` column is unique for seven key regions.
#'
#' @source https://api.worldbank.org/v2/region
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all regions
#' wdi_get_regions()
#'
wdi_get_regions <- function(language = "en") {
  regions_raw <- perform_request("region", language)

  if (!is.null(regions_raw)) {
    # id is non-missing for 7 entries
    regions_processed <- as_tibble(regions_raw) |>
      mutate(id = as.integer(.data$id)) |>
      select(
        region_id = "id",
        region_code = "code",
        region_iso2code = "iso2code",
        region_name = "name"
      ) |>
      mutate(across(where(is.character), trimws))

    regions_processed
  }
}

test_that("wdi_get_regions handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(result <- wdi_get_regions(language = "xx"))
  expect_null(result)
})

test_that("wdi_get_regions returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_regions()
  expect_s3_class(result, "tbl_df")
  expected_colnames <- c(
    "region_id",
    "region_code",
    "region_iso2code",
    "region_name"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_regions trims whitespace in character columns", {
  mock_data <- tibble(
    id = c(" 1 ", " 2 "),
    code = c(" EAS ", " ECS "),
    iso2code = c(" Z4 ", " Z7 "),
    name = c(" East Asia & Pacific ", " Europe & Central Asia ")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_regions()
      expect_equal(result$region_id, c(1, 2))
      expect_equal(result$region_code, c("EAS", "ECS"))
      expect_equal(result$region_iso2code, c("Z4", "Z7"))
      expect_equal(
        result$region_name,
        c("East Asia & Pacific", "Europe & Central Asia")
      )
    }
  )
})

test_that("wdi_get_regions converts id to integer", {
  mock_data <- tibble(
    id = c("1", "2"),
    code = c("EAS", "ECS"),
    iso2code = c("Z4", "Z7"),
    name = c("East Asia & Pacific", "Europe & Central Asia")
  )

  with_mocked_bindings(
    perform_request = function(endpoint, language) mock_data,
    {
      result <- wdi_get_regions()
      expect_type(result$region_id, "integer")
      expect_equal(result$region_id, c(1L, 2L))
    }
  )
})

test_that("wdi_get_regions handles empty data gracefully", {
  mock_data <- tibble(
    id = character(),
    code = character(),
    iso2code = character(),
    name = character()
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_regions()
      expect_equal(nrow(result), 0)
    }
  )
})

test_that("wdi_get_regions handles different language inputs", {
  skip_if_offline()

  result <- wdi_get_regions(language = "fr")
  expected_columns <- c(
    "region_id",
    "region_code",
    "region_iso2code",
    "region_name"
  )
  expect_true(all(expected_columns %in% colnames(result)))
})

#' Download lending types from the World Bank API
#'
#' This function returns a tibble of supported lending types for querying the
#' World Bank API. The lending types classify countries based on the financial
#' terms available to them from the World Bank.
#'
#' @param language A character string specifying the language code for the API
#'  response (default is "en" for English).
#'
#' @return A tibble with columns that typically include:
#' \describe{
#'   \item{lending_type_id}{An integer for the lending type.}
#'   \item{lending_type_iso2code}{A character string for the ISO2 code of the
#'                                lending type.}
#'   \item{lending_type_name}{A description of the lending type (e.g., "IBRD",
#'                            "IDA").}
#' }
#'
#' @details This function provides a reference for the supported lending types,
#' which classify countries according to the financial terms they are eligible
#' for under World Bank programs. The language parameter allows the results to
#' be returned in different languages as supported by the API.
#'
#' @source https://api.worldbank.org/v2/lendingTypes
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all lending types in English
#' wdi_get_lending_types()
#'
wdi_get_lending_types <- function(language = "en") {
  lending_types_raw <- perform_request("lendingTypes", language)

  if (!is.null(lending_types_raw)) {
    lending_types_processed <- as_tibble(lending_types_raw) |>
      select(
        lending_type_id = "id",
        lending_type_iso2code = "iso2code",
        lending_type_name = "value"
      ) |>
      mutate(across(where(is.character), trimws))

    lending_types_processed
  }
}

test_that("wdi_get_lending_types handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(result <- wdi_get_lending_types(language = "xx"))
  expect_null(result)
})

test_that("wdi_get_lending_types returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_lending_types()
  expect_s3_class(result, "tbl_df")
  expected_colnames <- c(
    "lending_type_id",
    "lending_type_iso2code",
    "lending_type_name"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_lending_types trims whitespace in character columns", {
  mock_data <- tibble(
    id = c(" IBRD ", " IDA "),
    iso2code = c(" B ", " D "),
    value = c(
      " International Bank for Reconstruction and Development ",
      " International Development Association "
    )
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_lending_types()
      expect_equal(result$lending_type_id, c("IBRD", "IDA"))
      expect_equal(result$lending_type_iso2code, c("B", "D"))
      expect_equal(
        result$lending_type_name,
        c(
          "International Bank for Reconstruction and Development",
          "International Development Association"
        )
      )
    }
  )
})

test_that("wdi_get_lending_types handles empty data gracefully", {
  mock_data <- tibble(
    id = character(),
    iso2code = character(),
    value = character()
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_lending_types()
      expect_equal(nrow(result), 0)
    }
  )
})

test_that("wdi_get_lending_types handles different language inputs", {
  skip_if_offline()

  result <- wdi_get_lending_types(language = "es")
  expected_colnames <- c(
    "lending_type_id",
    "lending_type_iso2code",
    "lending_type_name"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

#' Download all countries and regions from the World Bank API
#'
#' This function retrieves information about entities (countries and regions)
#' from the World Bank API. It returns a tibble containing various details such
#' as the entity's ID, ISO2 code, name, region information, lending type,
#' capital city, and coordinates.
#'
#' @param language A character string specifying the language for the API
#'  response. Defaults to `"en"` (English). Other supported options include
#'  `"es"` (Spanish), `"fr"` (French), and others depending on the API.
#' @param per_page An integer specifying the number of records to fetch per
#'  request. Defaults to `1000`.
#'
#' @return A tibble with the following columns:
#' \describe{
#'   \item{entity_id}{A character string representing the entity's unique
#'                       identifier.}
#'   \item{entity_name}{A character string for the name of the entity.}
#'   \item{entity_iso2code}{A character string for the ISO2 country code.}
#'   \item{entity_type}{A character string for the type of the entity
#'                         ("country" or "region").}
#'   \item{region_id}{A character string representing the region's unique
#'                    identifier.}
#'   \item{region_name}{A character string for the name of the region.}
#'   \item{region_iso2code}{A character string for the ISO2 region code.}
#'   \item{admin_region_id}{A character string representing the administrative
#'                          region's unique identifier.}
#'   \item{admin_region_name}{A character string for the name of the
#'                            administrative region.}
#'   \item{admin_region_iso2code}{A character string for the ISO2 code of the
#'                                administrative region.}
#'   \item{income_level_id}{A character string representing the entity's
#'                          income level.}
#'   \item{income_level_name}{A character string for the name of the
#'                            income level.}
#'   \item{income_level_iso2code}{A character string for the ISO2 code of the
#'                                income level.}
#'   \item{lending_type_id}{A character string representing the lending type's
#'                          unique identifier.}
#'   \item{lending_type_name}{A character string for the name of the lending
#'                            type.}
#'   \item{lending_type_iso2code}{A character string for the ISO2 code of the
#'                                lending type.}
#'   \item{capital_city}{A character string for the name of the capital city.}
#'   \item{longitude}{A numeric value for the longitude of the entity.}
#'   \item{latitude}{A numeric value for the latitude of the entity.}
#' }
#'
#' @details This function sends a request to the World Bank API to retrieve data
#' for all supported entities in the specified language. The data is then
#' processed into a tidy format and includes information about the country,
#' such as its ISO code, capital city, geographical coordinates, and additional
#' metadata about regions, income levels, and lending types.
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all entities in English
#' wdi_get_entities()
#'
#' # Download all entities in Spanish
#' wdi_get_entities(language = "zh")
#'
wdi_get_entities <- function(language = "en", per_page = 1000) {
  entities_raw <- perform_request("countries/all", language, per_page)

  if (!is.null(entities_raw)) {
    entities_processed <- entities_raw |>
      as_tibble() |>
      rename(
        entity_id = "id",
        entity_iso2code = "iso2Code",
        entity_name = "name"
      ) |>
      unnest_wider("region") |>
      rename(
        region_id = "id",
        region_iso2code = "iso2code",
        region_name = "value"
      ) |>
      unnest_wider("adminregion") |>
      rename(
        admin_region_id = "id",
        admin_region_iso2code = "iso2code",
        admin_region_name = "value"
      ) |>
      unnest_wider("incomeLevel") |>
      rename(
        income_level_id = "id",
        income_level_iso2code = "iso2code",
        income_level_name = "value"
      ) |>
      unnest_wider("lendingType") |>
      rename(
        lending_type_id = "id",
        lending_type_iso2code = "iso2code",
        lending_type_name = "value",
        capital_city = "capitalCity"
      ) |>
      mutate(
        longitude = if_else(
          .data$longitude == "",
          NA_real_,
          as.numeric(.data$longitude)
        ),
        latitude = if_else(
          .data$latitude == "",
          NA_real_,
          as.numeric(.data$latitude)
        ),
        entity_type = if_else(
          .data$region_name == "Aggregates",
          "aggregates",
          "country"
        ),
        across(where(is.character), ~ if_else(.x == "", NA, .x)),
        across(where(is.character), trimws)
      ) |>
      relocate(c("entity_type", "capital_city"), .after = "entity_iso2code")

    entities_processed
  }
}

test_that("wdi_get_entities handles invalid language input gracefully", {
  skip_if_offline()

  expect_message(result <- wdi_get_entities(language = "xx"))
  expect_null(result)
})

test_that("wdi_get_entities handles invalid per_page input", {
  expect_error(
    wdi_get_entities(per_page = -1),
  )
  expect_error(
    wdi_get_entities(per_page = 50000)
  )
  expect_error(
    wdi_get_entities(per_page = "500")
  )
})

test_that("wdi_get_entities returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_entities()
  expect_s3_class(result, "tbl_df")
  expected_colnames <- c(
    "entity_id",
    "entity_name",
    "entity_iso2code",
    "entity_type",
    "region_id",
    "region_name",
    "region_iso2code",
    "admin_region_id",
    "admin_region_name",
    "admin_region_iso2code",
    "income_level_id",
    "income_level_name",
    "income_level_iso2code",
    "lending_type_id",
    "lending_type_name",
    "lending_type_iso2code",
    "capital_city",
    "longitude",
    "latitude"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

test_that("wdi_get_entities handles type conversions and missing values", {
  mock_data <- tibble(
    id = "ABW",
    iso2Code = "AW",
    name = "Aruba",
    region = list(tibble(
      id = "LCN",
      iso2code = "ZJ",
      value = "Latin America & Caribbean"
    )),
    adminregion = list(tibble(
      id = NA_character_,
      iso2code = NA_character_,
      value = NA_character_
    )),
    incomeLevel = list(tibble(
      id = "HIC",
      iso2code = "XD",
      value = "High income"
    )),
    lendingType = list(tibble(id = "IBD", iso2code = "XF", value = "IBRD")),
    capitalCity = "Oranjestad",
    longitude = "70.0167",
    latitude = "12.5167"
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_entities()
      expect_equal(result$entity_id, "ABW")
      expect_equal(result$entity_iso2code, "AW")
      expect_equal(result$entity_name, "Aruba")
      expect_equal(result$region_id, "LCN")
      expect_equal(result$region_name, "Latin America & Caribbean")
      expect_equal(result$income_level_id, "HIC")
      expect_equal(result$lending_type_name, "IBRD")
      expect_equal(result$capital_city, "Oranjestad")
      expect_equal(result$longitude, 70.0167)
      expect_equal(result$latitude, 12.5167)
      expect_true(all(c("longitude", "latitude") %in% colnames(result)))
    }
  )
})

test_that("wdi_get_entities handles different language inputs", {
  skip_if_offline()

  result <- wdi_get_entities(language = "fr")

  expected_colnames <- c(
    "entity_id",
    "entity_name",
    "entity_iso2code",
    "entity_type",
    "region_id",
    "region_name",
    "region_iso2code",
    "admin_region_id",
    "admin_region_name",
    "admin_region_iso2code",
    "income_level_id",
    "income_level_name",
    "income_level_iso2code",
    "lending_type_id",
    "lending_type_name",
    "lending_type_iso2code",
    "capital_city",
    "longitude",
    "latitude"
  )
  expect_true(all(expected_colnames %in% colnames(result)))
})

#' Download languages from the World Bank API
#'
#' This function returns a tibble of supported languages for querying the World
#' Bank API. The supported languages include English, Spanish, French, Arabic,
#' and Chinese, etc.
#'
#' @return A tibble with three columns:
#' \describe{
#'   \item{language_code}{A character string representing the language code
#'                        (e.g., "en" for English).}
#'   \item{language_name}{A character string representing the description of the
#'                        language (e.g., "English").}
#'   \item{native_form}{A character string representing the native form of the
#'                      language (e.g., "English").}
#' }
#'
#' @details This function provides a simple reference for the supported
#' languages when querying the World Bank API.
#'
#' @source https://api.worldbank.org/v2/languages
#'
#' @export
#'
#' @examplesIf curl::has_internet()
#' # Download all languages
#' wdi_get_languages()
#'
wdi_get_languages <- function() {
  # languages resource does not support multiple languages
  languages_raw <- perform_request("languages")

  if (!is.null(languages_raw)) {
    languages_processed <- as_tibble(languages_raw) |>
      select(
        language_code = "code",
        language_name = "name",
        native_form = "nativeForm"
      ) |>
      mutate(across(where(is.character), trimws))

    languages_processed
  }
}

test_that("wdi_get_languages returns a tibble with correct column names", {
  skip_if_offline()

  result <- wdi_get_languages()
  expect_s3_class(result, "tbl_df")
  expected_columns <- c("language_code", "language_name", "native_form")
  expect_true(all(expected_columns %in% colnames(result)))
})

test_that("wdi_get_languages trims whitespace in character columns", {
  mock_data <- tibble(
    code = c(" en ", " es "),
    name = c(" English ", " Spanish "),
    nativeForm = c(" English ", " Español ")
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_languages()
      expect_equal(result$language_code, c("en", "es"))
      expect_equal(result$language_name, c("English", "Spanish"))
      expect_equal(result$native_form, c("English", "Español"))
    }
  )
})

test_that("wdi_get_languages returns NULL when API is unavailable", {
  with_mocked_bindings(
    perform_request = function(...) NULL,
    {
      result <- wdi_get_languages()
      expect_null(result)
    }
  )
})

test_that("wdi_get_languages handles empty data gracefully", {
  mock_data <- tibble(
    code = character(),
    name = character(),
    nativeForm = character()
  )

  with_mocked_bindings(
    perform_request = function(...) mock_data,
    {
      result <- wdi_get_languages()
      expect_equal(nrow(result), 0)
    }
  )
})

#' @keywords internal
#' @import httr2
#' @import dplyr
#' @import tibble
#' @import jsonlite
"_PACKAGE"

## usethis namespace: start
#' @importFrom purrr map map_df reduce
#' @importFrom tidyr unnest nest pivot_wider unnest_wider unnest_longer
#' @importFrom rlang parse_expr eval_tidy .data %||%
## usethis namespace: end
NULL

---
title: "`wbwdi` Data Model"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{`wbwdi` Data Model}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Overview

The diagram below shows the data model of the `wbwdi` package. In our design, we primarily have R users in mind, particularly those who rely heavily on the popular data manipulation packages `dplyr` and `data.table`. For these users, having consistent and descriptive primary key column names across different tables (e.g., `entity_id`, `series_id`) simplifies writing joins across tables, reduces the risk of column name conflicts, and avoids ambiguity. For more information on the design philosophy, we refer to [econdataverse.org](https://www.econdataverse.org/).

```{r, out.width = "100%", echo=FALSE, fig.alt = "A visual representation of the data model behind the 'wbwdi' R package."}
knitr::include_graphics("images/data-model.png")
```

<!-- ```{mermaid} -->
<!-- erDiagram -->
<!--   ENTITIES ||--o{ WORLD_DEVELOPMENT_INDICATORS : has -->
<!--   ENTITIES { -->
<!--     string entity_id PK -->
<!--     string entity_name -->
<!--     string entity_iso2code -->
<!--     string entity_type -->
<!--     string capital_city -->
<!--     string region_id -->
<!--     string region_name -->
<!--     string region_iso2code -->
<!--     string admin_region_id -->
<!--     string admin_region_name -->
<!--     string admin_region_iso2code -->
<!--     string income_level_id -->
<!--     string income_level_name -->
<!--     string income_level_iso2code -->
<!--     string lending_type_id -->
<!--     string lending_type_name -->
<!--     string lending_type_iso2code -->
<!--     float longitude -->
<!--     float latitude -->
<!--   } -->
<!--   INDICATORS ||--o{ WORLD_DEVELOPMENT_INDICATORS : has -->
<!--   INDICATORS { -->
<!--     string indicator_id PK -->
<!--     string indicator_name -->
<!--     int source_id FK -->
<!--     string source_name -->
<!--     string source_note -->
<!--     string source_organization -->
<!--     list topics -->
<!--   } -->
<!--   TOPICS ||--o{ INDICATORS : has -->
<!--   TOPICS { -->
<!--     int topic_id PK -->
<!--     string topic_name -->
<!--     string topic_note -->
<!--   } -->
<!--   SOURCES ||--o{ INDICATORS : has -->
<!--   SOURCES { -->
<!--     int source_id PK -->
<!--     string source_name -->
<!--     date update_date -->
<!--     logical is_data_available -->
<!--     logical is_metadata_available -->
<!--     int concepts -->
<!--   } -->
<!--   WORLD_DEVELOPMENT_INDICATORS { -->
<!--     string entity_id FK -->
<!--     string indicator_id FK -->
<!--     int year -->
<!--     int quarter -->
<!--     int month -->
<!--     float value -->
<!--   } -->
<!-- ``` -->

## Table details

### entities

| Column name | Description | Example value |
|------------------|------------------------------------|------------------|
| entity_id | Unique identifier for the entity | ZMB |
| entity_name | Standardized name of the entity | Zambia |
| entity_iso2code | ISO 3166-1 alpha-2 code of the entity | ZM |
| entity_type | Type of entity (country, aggregate) | country |
| capital_city | Capital city of the entity | Lusaka |
| region_id | Unique identifier for the region | SSF |
| region_name | Name of the region | Sub-Saharan Africa |
| region_iso2code | ISO 3166-1 alpha-2 code of the region | ZG |
| admin_region_id | Unique identifier for the administrative region | SSA |
| admin_region_name | Name of the administrative region | Sub-Saharan Africa (excluding high income) |
| admin_region_iso2code | Unique identifier for the administrative region | ZF |
| income_level_id | Unique identifier for the income level | LMC |
| income_level_name | Name of the income level | Lower middle income |
| income_level_iso2code | ISO code of the income level | XN |
| lending_type_id | Unique identifier for the lending type | IDX |
| lending_type_name | Name of the lending type | IDA |
| lending_type_iso2code | ISO code of the lending type | XI |
| longitude | Longitude of the entity | 28.2937 |
| latitude | Latitude of the entity | -15.3982 |

### Indicators

| Column name | Description | Example value |
|------------------|--------------------|----------------------------------|
| indicator_id | Unique identifier for the indicator | DT.DOD.DPPG.CD |
| indicator_name | Name of the indicator | External debt stocks, public and publicly guaranteed (PPG) (DOD, current US\$) |
| source_id | Unique identifier for the data source | 2 |
| source_name | Name for the data source | World Development Indicators |
| source_note | Note about the data source | Public and publicly guaranteed debt comprises long-term external obligations of public debtors, including the national government, Public Corporations, State Owned Enterprises, Development Banks and Other Mixed Enterprises, political subdivisions (or an agency of either), autonomous public bodies, and external obligations of private debtors that are guaranteed for repayment by a public entity. Data are in current U.S. dollars. |
| source_organization | Organization responsible for the data series | World Bank, International Debt Statistics. |
| topics | List of topics with ID and corresponding name | (20, External Debt) |

### Topics

| Column name | Description | Example value |
|------------------|--------------------|----------------------------------|
| topic_id | Unique identifier for the topic | 1 |
| topic_name | Name of the topic | Agriculture & Rural Development |
| topic_note | Note about the topic | For the 70 percent of the world's poor who live in rural areas, agriculture is the main source of income and employment. But depletion and degradation of land and water pose serious challenges to producing enough food and other agricultural products to sustain livelihoods here and meet the needs of urban populations. Data presented here include measures of agricultural inputs, outputs, and productivity compiled by the UN's Food and Agriculture Organization. |

### Sources

| Column name | Description | Example value |
|------------------|--------------------|----------------------------------|
| source_id | Unique identifier for the source | 1 |
| source_name | Name of the source | Doing Business |
| update_date | Date with last update | 2021-08-18 |
| is_data_available | Indicator whether data is available | TRUE |
| is_metadata_available | Indicator whether metadata is available | TRUE |
| concepts | Number of concepts in the source | 3 |

### World Development Indicators

| Column name  | Description                  | Example value  |
|--------------|------------------------------|----------------|
| entity_id | Identifier for the entity | USA            |
| indicator_id | Identifier for the indicator | DT.DOD.DPPG.CD |
| year         | Year of the data point       | 2020           |
| quarter      | Quarter of the data point (if quarterly data is requested and available).      | 3           |
| month         | Month of the data point (if monthly data is requested and available).    | 2020         1  |
| value        | Value of the data point      | 4298957000     |

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_indicators.R
\name{wdi_get_indicators}
\alias{wdi_get_indicators}
\title{Download all available World Bank indicators}
\usage{
wdi_get_indicators(language = "en", per_page = 32500L)
}
\arguments{
\item{language}{A character string specifying the language for the request,
see \link{wdi_get_languages}. Defaults to \code{"en"}.}

\item{per_page}{An integer specifying the number of results per page for the
API. Defaults to 32,500. Must be a value between 1 and 32,500.}
}
\value{
A tibble containing the available indicators and their metadata:
\describe{
\item{indicator_id}{A character string for the ID of the indicator (e.g.,
"NY.GDP.PCAP.KD").}
\item{indicator_name}{A character string for the name of the indicator
(e.g., "GDP per capita, constant prices").}
\item{source_id}{An integer identifying the data source providing the
indicator.}
\item{source_name}{A character string describing the source of the
indicator data.}
\item{source_note}{A character string providing additional notes about the
data source.}
\item{source_organization}{A character string denoting the organization
responsible for the data source.}
\item{topics}{A nested tibble containing (possibly multiple) topics
associated with the indicator, with two columns: an integer
\code{topic_id} and a character \code{topic_name}.}
}
}
\description{
This function retrieves a comprehensive list of all indicators supported by
the World Bank API. The indicators include metadata such as the indicator ID,
name, unit, source, and associated topics. The user can specify the language
of the API response and whether to include additional details.
}
\details{
This function makes a request to the World Bank API to retrieve
metadata for all available indicators. It processes the response into a tidy
tibble format.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# Download all supported indicators in English
wdi_get_indicators()

# Download all supported indicators in Spanish
wdi_get_indicators(language = "es")
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_regions.R
\name{wdi_get_regions}
\alias{wdi_get_regions}
\title{Download regions from the World Bank API}
\source{
https://api.worldbank.org/v2/region
}
\usage{
wdi_get_regions(language = "en")
}
\arguments{
\item{language}{A character string specifying the language code for the API
response (default is "en" for English).}
}
\value{
A tibble with the following columns:
\describe{
\item{region_id}{An integer identifier for each region.}
\item{region_code}{A character string representing the region code.}
\item{region_iso2code}{A character string representing the ISO2 code for
the region.}
\item{region_name}{A character string representing the name of the region,
in the specified language.}
}
}
\description{
This function returns a tibble of supported regions for querying the World
Bank API.The regions include various geographic areas covered by the World
Bank's datasets.
}
\details{
This function provides a reference for the supported regions, which
are important for refining queries related to geographic data in the World
Bank's datasets. The \code{region_id} column is unique for seven key regions.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all regions
wdi_get_regions()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_entities.R
\name{wdi_get_entities}
\alias{wdi_get_entities}
\title{Download all countries and regions from the World Bank API}
\usage{
wdi_get_entities(language = "en", per_page = 1000)
}
\arguments{
\item{language}{A character string specifying the language for the API
response. Defaults to \code{"en"} (English). Other supported options include
\code{"es"} (Spanish), \code{"fr"} (French), and others depending on the API.}

\item{per_page}{An integer specifying the number of records to fetch per
request. Defaults to \code{1000}.}
}
\value{
A tibble with the following columns:
\describe{
\item{entity_id}{A character string representing the entity's unique
identifier.}
\item{entity_name}{A character string for the name of the entity.}
\item{entity_iso2code}{A character string for the ISO2 country code.}
\item{entity_type}{A character string for the type of the entity
("country" or "region").}
\item{region_id}{A character string representing the region's unique
identifier.}
\item{region_name}{A character string for the name of the region.}
\item{region_iso2code}{A character string for the ISO2 region code.}
\item{admin_region_id}{A character string representing the administrative
region's unique identifier.}
\item{admin_region_name}{A character string for the name of the
administrative region.}
\item{admin_region_iso2code}{A character string for the ISO2 code of the
administrative region.}
\item{income_level_id}{A character string representing the entity's
income level.}
\item{income_level_name}{A character string for the name of the
income level.}
\item{income_level_iso2code}{A character string for the ISO2 code of the
income level.}
\item{lending_type_id}{A character string representing the lending type's
unique identifier.}
\item{lending_type_name}{A character string for the name of the lending
type.}
\item{lending_type_iso2code}{A character string for the ISO2 code of the
lending type.}
\item{capital_city}{A character string for the name of the capital city.}
\item{longitude}{A numeric value for the longitude of the entity.}
\item{latitude}{A numeric value for the latitude of the entity.}
}
}
\description{
This function retrieves information about entities (countries and regions)
from the World Bank API. It returns a tibble containing various details such
as the entity's ID, ISO2 code, name, region information, lending type,
capital city, and coordinates.
}
\details{
This function sends a request to the World Bank API to retrieve data
for all supported entities in the specified language. The data is then
processed into a tidy format and includes information about the country,
such as its ISO code, capital city, geographical coordinates, and additional
metadata about regions, income levels, and lending types.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all entities in English
wdi_get_entities()

# Download all entities in Spanish
wdi_get_entities(language = "zh")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wbwdi-package.R
\docType{package}
\name{wbwdi-package}
\alias{wbwdi}
\alias{wbwdi-package}
\title{wbwdi: Seamless Access to World Bank World Development Indicators (WDI)}
\description{
Access and analyze the World Bank's World Development Indicators (WDI) using the corresponding API \url{https://datahelpdesk.worldbank.org/knowledgebase/articles/889392-about-the-indicators-api-documentation}. WDI provides more than 24,000 country or region-level indicators for various contexts. 'wbwdi' enables users to download, process and work with WDI series across multiple countries, aggregates, and time periods.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/tidy-intelligence/r-wbwdi}
  \item \url{https://tidy-intelligence.github.io/r-wbwdi/}
  \item Report bugs at \url{https://github.com/tidy-intelligence/r-wbwdi/issues}
}

}
\author{
\strong{Maintainer}: Christoph Scheuch \email{christoph@tidy-intelligence.com} (\href{https://orcid.org/0009-0004-0423-6819}{ORCID}) [copyright holder]

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/perform_request.R
\name{perform_request}
\alias{perform_request}
\title{Internal function to perform API requests to the World Bank API}
\usage{
perform_request(
  resource,
  language = NULL,
  per_page = 10000L,
  date = NULL,
  most_recent_only = NULL,
  source = NULL,
  progress = FALSE,
  base_url = "https://api.worldbank.org/v2/",
  max_tries = 10L
)
}
\arguments{
\item{resource}{A character string specifying the resource endpoint for the
World Bank API (e.g., "incomeLevels", "lendingTypes").}

\item{language}{A character string specifying the language code for the API
response. Defaults to NULL.}

\item{per_page}{An integer specifying the number of results per page for the
API. Defaults to 10,000.
Must be a value between 1 and 32,500.}

\item{date}{A character string specifying the date range of the data to be
retrieved (e.g., "2000:2020"). If NULL (default), no date filtering is
applied.}

\item{source}{A character string specifying the data source for the API
request. If NULL (default),
no specific source is selected.}

\item{progress}{A logical value indicating whether to display a progress bar
for paginated requests. Defaults to FALSE.}

\item{base_url}{A character string specifying the base URL of the World Bank
API (default is "https://api.worldbank.org/v2").}

\item{max_tries}{An integer specifying the number of attempts for the API
request.}
}
\value{
A list containing the parsed JSON response from the API. If the
response contains multiple pages, the results are combined into a single
list. If the API returns an error, it provides the relevant error message
and documentation.
}
\description{
This function constructs and sends a request to the specified World Bank API
resource, with options to specify the language, the number of results per
page, a date range, and the data source. It also supports paginated requests
and progress tracking.
}
\details{
This internal helper function constructs the request URL using the
base URL, language, resource, and other optional parameters (such as date and
source), and performs the API request. For paginated results, it iterates
through all pages and consolidates the data into a single output. The
function handles API errors, providing detailed error messages when
available.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_income_levels.R
\name{wdi_get_income_levels}
\alias{wdi_get_income_levels}
\title{Download income levels from the World Bank API}
\source{
https://api.worldbank.org/v2/incomeLevels
}
\usage{
wdi_get_income_levels(language = "en")
}
\arguments{
\item{language}{A character string specifying the language code for the API
response (default is "en" for English).}
}
\value{
A tibble with columns that typically include:
\describe{
\item{income_level_id}{An integer identifier for the income level.}
\item{income_level_iso2code}{A character string representing the ISO2 code
for the income level.}
\item{income_level_name}{The description of the income level (e.g.,
"Low income", "High income").}
}
}
\description{
This function returns a tibble of supported income levels for querying the
World Bank API. The income levels categorize countries based on their gross
national income per capita.
}
\details{
This function provides a reference for the supported income levels,
which categorize countries according to their income group as defined by the
World Bank. The language parameter allows the results to be returned in
different languages as supported by the API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all income levels in English
wdi_get_income_levels()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_lending_types.R
\name{wdi_get_lending_types}
\alias{wdi_get_lending_types}
\title{Download lending types from the World Bank API}
\source{
https://api.worldbank.org/v2/lendingTypes
}
\usage{
wdi_get_lending_types(language = "en")
}
\arguments{
\item{language}{A character string specifying the language code for the API
response (default is "en" for English).}
}
\value{
A tibble with columns that typically include:
\describe{
\item{lending_type_id}{An integer for the lending type.}
\item{lending_type_iso2code}{A character string for the ISO2 code of the
lending type.}
\item{lending_type_name}{A description of the lending type (e.g., "IBRD",
"IDA").}
}
}
\description{
This function returns a tibble of supported lending types for querying the
World Bank API. The lending types classify countries based on the financial
terms available to them from the World Bank.
}
\details{
This function provides a reference for the supported lending types,
which classify countries according to the financial terms they are eligible
for under World Bank programs. The language parameter allows the results to
be returned in different languages as supported by the API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all lending types in English
wdi_get_lending_types()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_search.R
\name{wdi_search}
\alias{wdi_search}
\title{Search for Keywords in Data Frame Columns}
\usage{
wdi_search(data, keywords, columns = NULL)
}
\arguments{
\item{data}{A data frame to search. It can include nested columns (lists
within columns).}

\item{keywords}{A character vector of keywords to search for within the
specified columns. The search is case-insensitive.}

\item{columns}{Optional. A character vector of column names to limit the
search to specific columns. If \code{NULL}, all columns in \code{data} will be
searched.}
}
\value{
A data frame containing only the rows where at least one keyword is
found in the specified columns. The returned data frame has the same
structure as \code{data}.
}
\description{
This function searches for specified keywords across columns in a data frame
and returns only the rows where at least one keyword appears in any of the
specified columns. It supports nested columns (lists within columns) and
provides the option to limit the search to specific columns.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# Download indicators
indicators <- wdi_get_indicators()

# Search for keywords "inequality" or "gender" across all columns
wdi_search(
  indicators,
  keywords = c("inequality", "gender")
)

# Search for keywords only within the "indicator_name" column
wdi_search(
  indicators,
  keywords = c("inequality", "gender"),
  columns = c("indicator_name")
)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_sources.R
\name{wdi_get_sources}
\alias{wdi_get_sources}
\title{Download data sources from the World Bank API}
\source{
https://api.worldbank.org/v2/sources
}
\usage{
wdi_get_sources(language = "en")
}
\arguments{
\item{language}{A character string specifying the language code for the API
response (default is "en" for English).}
}
\value{
A tibble with six columns:
\describe{
\item{source_id}{An integer identifier for the data source.}
\item{source_code}{A character string for the source code.}
\item{source_name}{The name of the data source (e.g., "World Development
Indicators").}
\item{update_date}{The last update date of the data source.}
\item{is_data_available}{A boolean indicating whether data is available.}
\item{is_metadata_available}{A boolean indicating whether metadata is
available.}
\item{concepts}{The number of concepts defined for the data source.}
}
}
\description{
This function returns a tibble of supported data sources for querying the
World Bank API. The data sources include various databases and datasets
provided by the World Bank.
}
\details{
This function provides a reference for the supported data sources
and their metadata when querying the World Bank API. The columns
\code{is_data_available} and \code{is_metadata_available} are logical values
derived from the API response, where "Y" indicates availability.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all available data sources in English
wdi_get_sources()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_languages.R
\name{wdi_get_languages}
\alias{wdi_get_languages}
\title{Download languages from the World Bank API}
\source{
https://api.worldbank.org/v2/languages
}
\usage{
wdi_get_languages()
}
\value{
A tibble with three columns:
\describe{
\item{language_code}{A character string representing the language code
(e.g., "en" for English).}
\item{language_name}{A character string representing the description of the
language (e.g., "English").}
\item{native_form}{A character string representing the native form of the
language (e.g., "English").}
}
}
\description{
This function returns a tibble of supported languages for querying the World
Bank API. The supported languages include English, Spanish, French, Arabic,
and Chinese, etc.
}
\details{
This function provides a simple reference for the supported
languages when querying the World Bank API.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all languages
wdi_get_languages()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get_topics.R
\name{wdi_get_topics}
\alias{wdi_get_topics}
\title{Download topics from the World Bank API}
\source{
https://api.worldbank.org/v2/topics
}
\usage{
wdi_get_topics(language = "en")
}
\arguments{
\item{language}{A character string specifying the language code for the API
response (default is "en" for English).}
}
\value{
A tibble with three columns:
\describe{
\item{id}{The unique identifier for the topic.}
\item{value}{The name of the topic (e.g., "Education", "Health").}
\item{source_note}{A brief description or note about the topic.}
}
}
\description{
This function returns a tibble of supported topics for querying the World
Bank API. Topics represent the broad subject areas covered by the World
Bank's datasets.
}
\details{
This function provides a reference for the supported topics that can
be used to refine your queries when accessing the World Bank API. Topics
represent different areas of focus for data analysis.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
# Download all available topics in English
wdi_get_topics()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wdi_get.R
\name{wdi_get}
\alias{wdi_get}
\title{Download World Bank indicator data for specific entities and time periods}
\usage{
wdi_get(
  entities,
  indicators,
  start_year = NULL,
  end_year = NULL,
  most_recent_only = FALSE,
  frequency = "annual",
  language = "en",
  per_page = 10000L,
  progress = TRUE,
  source = NULL,
  format = "long"
)
}
\arguments{
\item{entities}{A character vector of ISO 2 or ISO 3-country codes, or
\code{"all"} to retrieve data for all entities.}

\item{indicators}{A character vector specifying one or more World Bank
indicators to download (e.g., c("NY.GDP.PCAP.KD", "SP.POP.TOTL")).}

\item{start_year}{Optional integer. The starting date for the data as a year.}

\item{end_year}{Optional integer. The ending date for the data as a year.}

\item{most_recent_only}{A logical value indicating whether to download only
the most recent value. In case of \code{TRUE}, it overrides \code{start_year} and
\code{end_year}. Defaults to \code{FALSE}.}

\item{frequency}{A character string specifying the frequency of the data
("annual", "quarter", "month"). Defaults to "annual".}

\item{language}{A character string specifying the language for the request,
see \link{wdi_get_languages}. Defaults to \code{"en"}.}

\item{per_page}{An integer specifying the number of results per page for the
API. Defaults to 10,000.}

\item{progress}{A logical value indicating whether to show progress messages
during the data download and parsing. Defaults to \code{TRUE}.}

\item{source}{An integer value specifying the data source, see
\link{wdi_get_sources}.}

\item{format}{A character value specifying whether the data is returned in
\code{"long"} or \code{"wide"} format. Defaults to \code{"long"}.}
}
\value{
A tibble with the following columns:
\describe{
\item{entity_id}{The ISO 3-country code of the country or aggregate for
which the data was retrieved.}
\item{indicator_id}{The ID of the indicator (e.g., "NY.GDP.PCAP.KD").}
\item{year}{The year of the indicator data as an integer.}
\item{quarter}{Optional. The quarter of the indicator data as integer.}
\item{month}{Optional. The month of the indicator data as integer.}
\item{value}{The value of the indicator for the given country and date.}
}
}
\description{
This function retrieves indicator data from the World Bank API for a
specified set of entities and indicators. The user can specify one or more
indicators, a date range, and other options to tailor the request. The data
is processed and returned in a tidy format, including country, indicator,
date, and value fields.
}
\details{
This function constructs a request URL for the World Bank API,
retrieves the relevant data for the given entities and indicators, and
processes the response into a tidy format. The user can optionally specify a
date range, and the function will handle requests for multiple pages if
necessary. If the \code{progress} parameter is \code{TRUE}, messages will be displayed
during the request and parsing process.

The function supports downloading multiple indicators by sending individual
API requests for each indicator and then combining the results into a single
tidy data frame.
}
\examples{
\dontshow{if (curl::has_internet()) withAutoprint(\{ # examplesIf}
\donttest{
# Download single indicator for multiple entities
wdi_get(c("USA", "CAN", "GBR"), "NY.GDP.PCAP.KD")

# Download single indicator for a specific time frame
wdi_get(c("USA", "CAN", "GBR"), "DPANUSSPB",
        start_year = 2012, end_year = 2013)

# Download single indicator for monthly frequency
wdi_get("AUT", "DPANUSSPB",
        start_year = 2012, end_year = 2015, frequency = "month")

# Download single indicator for quarterly frequency
wdi_get("NGA", "DT.DOD.DECT.CD.TL.US",
        start_year = 2012, end_year = 2015, frequency = "quarter")

# Download single indicator for all entities and disable progress bar
wdi_get("all", "NY.GDP.PCAP.KD", progress = FALSE)

# Download multiple indicators for multiple entities
wdi_get(c("USA", "CAN", "GBR"), c("NY.GDP.PCAP.KD", "SP.POP.TOTL"))

# Download indicators for different sources
wdi_get("DEU", "NY.GDP.PCAP.KD", source = 2)
wdi_get("DEU", "SG.LAW.INDX", source = 14)

# Download indicators in wide format
wdi_get(c("USA", "CAN", "GBR"), c("NY.GDP.PCAP.KD"),
        format = "wide")
wdi_get(c("USA", "CAN", "GBR"), c("NY.GDP.PCAP.KD", "SP.POP.TOTL"),
        format = "wide")

# Download most recent value only
wdi_get("USA", "SP.POP.TOTL", most_recent_only = TRUE)
}
\dontshow{\}) # examplesIf}
}
